import time
from pathlib import Path
from typing import Optional
from .conf import get_setting


def _get_lock_path(lockname: str) -> Path:
    cache_dir = get_setting("caerp.static_tmp", default="/tmp")
    directory = Path(cache_dir)
    return directory.joinpath(f"{lockname}.lock")


def _get_file_age(path: Path) -> int:
    return int(time.time() - path.stat().st_mtime)


def is_locked(lockname, max_age: Optional[int] = None):
    """
    lockname: the name of the lock
    max_age: the maximum age of the lock file in seconds
    """
    p = _get_lock_path(lockname)
    result = p.exists()
    if result and max_age is not None and _get_file_age(p) > max_age:
        p.unlink()
        result = False
    return result


def acquire_lock(lockname):
    p = _get_lock_path(lockname)
    p.touch()


def release_lock(lockname):
    p = _get_lock_path(lockname)
    p.unlink()



import datetime
from openpyxl import load_workbook
import logging
from typing import Iterable, Optional, Generator

from pathlib import Path
from pyramid.request import Request

from caerp.compute.math_utils import str_to_int, str_to_float
from caerp.utils.datetimes import str_to_date

from caerp_celery.interfaces import IAccountingFileParser
from caerp_celery.parsers import BaseParser, BaseProducer, OperationData


logger = logging.getLogger(__name__)


class QuadraFileParser(BaseParser):
    """
    Parse un fichier xlsx extrait de quadra
    """

    def _stream_operation_lines_from_file(self) -> Generator[dict, None, None]:
        wb = load_workbook(self.file_path)
        ws = wb.active
        rows: Generator = ws.rows
        headers = [cell.value for cell in next(rows)]
        for row in rows:
            row_data = {}
            for num_col, header in enumerate(headers):
                row_data[header] = row[num_col].value
            yield row_data

    def stream(self) -> Generator[dict, None, None]:
        for line in self._stream_operation_lines_from_file():
            yield line


class OperationProducer(BaseProducer):
    def _get_num_val(self, line: dict, key: str) -> float:
        result = 0
        val = line.get(key, 0)
        result = val or 0
        if not isinstance(result, (float, int)):
            result = str_to_float(result, default=0)
        return result

    def _get_label(self, line: dict) -> str:
        label = line.get("Libellé", "").strip()
        label = label[:80]
        return label

    def _get_date(self, line: dict) -> datetime.date:
        day = str_to_int(line.get("Jour", "0"), default=0)
        date = str_to_date(line.get("Période", "").strip()).date()
        return datetime.date(year=date.year, month=date.month, day=day)

    def _stream_operation(self, line) -> Optional[OperationData]:
        for key in ("CentreSimple", "Compte", "Débit", "Libellé", "Crédit", "Période"):
            if key not in line.keys():
                logger.error("This line has incorrect datas : %s" % line)
                return None
        analytical_account = line.get("CentreSimple", "").strip()
        general_account = line.get("Compte", "").strip()
        date = self._get_date(line)
        if not date:
            logger.error("This line has incorrect date data : %s" % line)
            return None

        label = self._get_label(line)
        credit = self._get_num_val(line, key="Crédit")
        debit = self._get_num_val(line, key="Débit")
        balance = 0
        result = OperationData(
            analytical_account=analytical_account,
            general_account=general_account,
            date=date,
            label=label,
            debit=debit,
            credit=credit,
            balance=balance,
        )
        return result

    def stream_operations(self) -> Iterable[OperationData]:
        for line in self.parser.stream():
            data = self._stream_operation(line)
            if data is not None:
                yield data


def parser_factory(context: Path, request: Request):
    return QuadraFileParser(context)


def producer_factory(context: IAccountingFileParser, request: Request):
    return OperationProducer(context)



import csv
import logging
import datetime
from typing import Iterable, List, Optional, Generator

from pathlib import Path
from sylk_parser import SylkParser
from pyramid.request import Request

from caerp.utils.datetimes import str_to_date
from caerp.compute.math_utils import str_to_float

from caerp_celery.interfaces import IAccountingFileParser
from caerp_celery.parsers import BaseParser, BaseProducer, OperationData

logger = logging.getLogger(__name__)


class SlkFileParser(BaseParser):
    def stream(self) -> Generator[list, None, None]:
        for line in SylkParser(str(self.file_path)):
            yield line


class CsvFileParser(BaseParser):
    encoding = "utf-8"

    def stream(self) -> Generator[list, None, None]:
        with open(self.file_path, encoding=self.encoding) as fbuf:
            dialect = csv.Sniffer().sniff(fbuf.read())
            fbuf.seek(0)
            for line in csv.reader(fbuf, dialect):
                yield line


class OperationProducer(BaseProducer):
    def _get_num_val(self, line: List[str], index: int) -> float:
        result = 0
        if len(line) > index:
            result = line[index].strip() or 0
        return str_to_float(result)

    def _get_label(self, line: List[str]) -> str:
        label = line[5].strip()
        label = label[:80]
        return label

    def _get_date(self, line) -> Optional[datetime.date]:
        date = None
        date_str = line[2].strip()
        if date_str:
            try:
                date = str_to_date(line[2].strip()).date()
            except Exception:
                logger.error("This line has incorrect datas : %s" % line)
                date = None
        return date

    def _stream_operation(self, line) -> Optional[OperationData]:
        """
        Analyticial/General/date/.../.../label/débit/crédit
        """
        result = None
        if len(line) >= 6:
            if line[0].strip() not in (
                "Compte analytique de l'entrepreneur",
                "Numéro analytique",
                " analytique",
                "",
            ):
                analytical_account = line[0].strip()
                general_account = line[1].strip()
                date = str_to_date(line[2].strip()).date()
                if not date:
                    return None

                label = self._get_label(line)
                debit = self._get_num_val(line, index=6)
                credit = self._get_num_val(line, index=7)
                balance = 0
                result = OperationData(
                    analytical_account=analytical_account,
                    general_account=general_account,
                    date=date,
                    label=label,
                    debit=debit,
                    credit=credit,
                    balance=balance,
                )
        return result

    def stream_operations(self) -> Iterable[OperationData]:
        for line in self.parser.stream():
            data = self._stream_operation(line)
            if data is not None:
                yield data


def parser_factory(context: Path, request: Request):
    if context.suffix == ".slk":
        return SlkFileParser(context)
    elif context.suffix == ".csv":
        return CsvFileParser(context)


def producer_factory(context: IAccountingFileParser, request: Request):
    return OperationProducer(context)



from zope.interface import Interface, Attribute


class IAccountingFileParser(Interface):
    """
    Interface pour les outils de lecture de Grand Livre
    """

    def __init__(self, pyramid_request, file_path):
        """ """


class IAccountingOperationProducer(Interface):
    """
    Produces accounting line data

    From a file or other source (api for example)
    """

    def set_data_source(self, source):
        """
        Set the data source used to produce accounting lines

        source can be :
            - a File buffer (result of open(filepath))
            - api connection data (no implementation yet)
        """

    def stream_lines(self) -> dict:
        """yield accounting operations in dict format"""



"""
caerp related hacks used to ensure code compatibility from within celery tasks
"""
from pyramid.security import remember
from pyramid_layout.config import create_layout_manager
from caerp.views.render_api import Api


class DummyEvent:
    def __init__(self, request, context):
        self.request = request
        self.request.context = context
        self.context = context


def setup_rendering_hacks(pyramid_request, context):
    event = DummyEvent(pyramid_request, context)
    create_layout_manager(event)
    pyramid_request.api = Api(context, pyramid_request)



"""
Celery tasks used to asynchronously generate exports (like excel exports)


Workflow :
    user provide filters
    TODO : user provide columns

    For UserDatas exports, we need to add some fields


    1- Task entry
    2- retrieve model
    3- generate the file or re-use the cached one
"""
import json
import os
import transaction
from tempfile import mktemp

from dateutil.utils import today
from pyramid_celery import celery_app
from beaker.cache import cache_region
from celery.utils.log import get_task_logger

from sqla_inspect.ods import SqlaOdsExporter, OdsExporter
from sqla_inspect.excel import SqlaXlsExporter, XlsExporter
from sqla_inspect.csv import SqlaCsvExporter, CsvExporter
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql.expression import label
from sqlalchemy import desc, func

from caerp.models.company import Company
from caerp.utils.datetimes import format_date
from caerp_celery.conf import get_setting, get_request
from caerp_celery.models import FileGenerationJob
from caerp_celery.tasks import utils
from caerp.compute.math_utils import round, integer_to_amount
from caerp.export.utils import format_filename
from caerp.export.sale_product import serialize_catalog
from caerp.models.task import Task
from caerp.models.expense import ExpenseSheet
from caerp.utils.html import strip_html_tags


MODELS_CONFIGURATION = {}

logger = utils.get_logger(__name__)


GENERATION_ERROR_MESSAGE = (
    "Une erreur inconnue a été rencontrée à la génération de votre fichier, "
    "veuillez contacter votre administrateur en lui "
    "fournissant l'identifiant suivant : %s"
)


def _add_o2m_headers_to_writer(writer, query, id_key):
    """
    Add column headers in the form "label 1",  "label 2" ... to be able to
    insert the o2m related elements to a main model's table export (allow to
    have 3 dimensionnal datas in a 2d array)

    E.g : Userdatas objects have got a o2m relationship on DateDatas objects

    Here we would add date 1, date 2... columns regarding the max number of
    configured datas (if a userdatas has 5 dates, we will have 5 columns)
    We fill the column with the value of an attribute of the DateDatas model
    (that is handled by sqla_inspect thanks to the couple index + related_key
    configuration)

    The name of the attribute is configured using the "flatten" key in the
    relationship's export configuration

    :param str id_key: The foreign key attribute mostly matching the class we
    export (e.g : when exporting UserDatas, most of the related elements point
    to it through a userdatas_id foreign key)
    """
    from caerp_base.models.base import DBSESSION

    new_headers = []
    for header in writer.headers:
        if isinstance(header["__col__"], RelationshipProperty):
            if header["__col__"].uselist:
                class_ = header["__col__"].mapper.class_
                # On compte le nombre maximum d'objet lié que l'on rencontre
                # dans la base
                if not hasattr(class_, id_key):
                    continue

                count = (
                    DBSESSION()
                    .query(label("nb", func.count(class_.id)))
                    .group_by(getattr(class_, id_key))
                    .order_by(desc("nb"))
                    .first()
                )

                if count is not None:
                    count = count[0]
                else:
                    count = 0

                # Pour les relations O2M qui ont un attribut flatten de
                # configuré, On rajoute des colonnes "date 1" "date 2" dans
                # notre sheet principale
                for index in range(0, count):
                    if "flatten" in header:
                        flatten_keys = header["flatten"]
                        if not hasattr(flatten_keys, "__iter__"):
                            flatten_keys = [flatten_keys]

                        for flatten_key, flatten_label in flatten_keys:
                            new_header = {
                                "__col__": header["__col__"],
                                "label": "%s %s %s"
                                % (header["label"], flatten_label, index + 1),
                                "key": header["key"],
                                "name": "%s_%s_%s"
                                % (header["name"], flatten_key, index + 1),
                                "related_key": flatten_key,
                                "index": index,
                            }
                            new_headers.append(new_header)

    writer.headers.extend(new_headers)
    return writer


def _get_tmp_directory_path():
    """
    Return the tmp filepath configured in the current configuration
    :param obj request: The pyramid request object
    """
    asset_path_spec = get_setting("caerp.static_tmp", mandatory=True)
    return asset_path_spec


def _get_tmp_filepath(directory, basename, extension):
    """
    Return a temp filepath for the given filename

    :param str basename: The base name to use
    :returns: A path to a non existing file
    :rtype: str
    """
    if not extension.startswith("."):
        extension = "." + extension

    filepath = mktemp(prefix=basename, suffix=extension, dir=directory)
    while os.path.exists(filepath):
        filepath = mktemp(prefix=basename, suffix=extension, dir=directory)
    return filepath


def _get_open_file(filepath, extension):
    """
    Get the appropriate writing mode regarding the provided extension
    """
    if extension in ("csv", "json"):
        return open(filepath, "w", newline="")
    else:
        return open(filepath, "wb")


def _transform_list_header_in_export_header(list_header, col_prefix="col_"):
    """
    Transform header formated in list of lists
    (eg: ['aaa', 'bbb', 'ccc'])
    into an exportable list of dicts
    (eg: [
            {"label": "aaa", "name": "col_1"},
            {"label": "bbb", "name": "col_2"},
            {"label": "ccc", "name": "col_3"},
        ]
    )
    """
    col_num = 1
    export_header = []
    for header in list_header:
        export_header.append(
            {
                "label": header,
                "name": f"{col_prefix}{col_num}",
            }
        )
        col_num += 1
    return export_header


def _transform_list_data_in_export_data(list_data, col_prefix="col_"):
    """
    Transform data formated in list of lists
    (eg: [['aaa', 'bbb', 'ccc'], ['ddd', 'eee', 'fff']])
    into an exportable list of dicts
    (eg: [
            {"col_1": "aaa", "col_2": "bbb", "col_3": "ccc"},
            {"col_1": "ddd", "col_2": "eee", "col_3": "fff"},
        ]
    )
    """
    export_data = []
    if len(list_data) > 0:
        col_num = 1
        col_list = []
        while col_num <= len(list_data[0]):
            col_list.append(f"{col_prefix}{col_num}")
            col_num += 1
        for row in list_data:
            export_data.append(dict(zip(col_list, row)))
    return export_data


@cache_region("default_term")
def _write_file_on_disk(tmpdir, model_type, ids, filename, extension):
    """
    Return a path to a generated file

    :param str tmpdir: The path to write to
    :param str model_type: The model key we want to generate an ods file for
    :param list ids: An iterable containing all ids of models to be included in
    the output
    :param str filename: The path to the file output
    :param str extension: The desired extension (xls/ods)
    :returns: The name of the generated file (unique and temporary name)
    :rtype: str
    """
    logger.debug(" No file was cached yet")
    config = MODELS_CONFIGURATION[model_type]
    model = config["factory"]
    query = model.query()
    if ids is not None:
        query = query.filter(model.id.in_(ids))

    options = {}
    if "excludes" in config:
        options["excludes"] = config["excludes"]
    if "order" in config:
        options["order"] = config["order"]
    if extension == "ods":
        writer = SqlaOdsExporter(model=model, **options)
    elif extension in ("xls", "xlsx"):
        writer = SqlaXlsExporter(model=model, **options)
    elif extension == "csv":
        writer = SqlaCsvExporter(model=model, **options)

    writer = _add_o2m_headers_to_writer(writer, query, config["foreign_key_name"])

    if "hook_init" in config:
        writer = config["hook_init"](writer, query)

    for item in query:
        writer.add_row(item)
        if "hook_add_row" in config:
            config["hook_add_row"](writer, item)

    filepath = _get_tmp_filepath(tmpdir, filename, extension)
    logger.debug(" + Writing file to %s" % filepath)

    # Since csv module expects strings
    with _get_open_file(filepath, extension) as f_buf:
        writer.render(f_buf)
    return os.path.basename(filepath)


@celery_app.task(bind=True)
def export_to_file(self, job_id, model_type, ids, filename="test", file_format="ods"):
    """
    Export the datas provided in the given query to ods format and generate a

    :param int job_id: The id of the job object used to record file_generation
    informations
    :param str model_type: The model we want to export (see MODELS)
    :param list ids: List of ids to query
    :param str filename: The base filename to use for the export (unique string
    is appended)
    :param str file_format: The format in which we want to export
    """
    logger = get_task_logger(__name__)
    logger.info("Exporting to a file")
    logger.info(" + model_type : %s", model_type)
    logger.info(" + ids : %s", ids)

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    filename = format_filename(filename)

    # Execute actions
    try:
        tmpdir = _get_tmp_directory_path()
        result_filename = _write_file_on_disk(
            tmpdir,
            model_type,
            ids,
            filename,
            file_format,
        )
        logger.debug(" -> The file %s been written", result_filename)
        transaction.commit()
    except Exception:
        transaction.abort()
        logger.exception("Error while generating file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)
    else:
        utils.record_completed(FileGenerationJob, job_id, filename=result_filename)

    return ""


def _export_data_to_tabular_file(headers, data, filename="export", file_format="csv"):
    """
    Generic function to export given data to given format

    :param list headers: List of dicts with headers data, format :
        (
            {"label": "Column label", "name": "col_1"},
            {"label": "Column label 2", "name": "col_2"},
        )
    :param list data: List of dicts with rows data for each headers, format :
        (
            {"col_1": "Value row 1 col 1", "col_2": "Value row 1 col 2"},
            {"col_1": "Value row 2 col 1", "col_2": "Value row 2 col 2"},
        )
    :param str filename: The base filename to use for the export
    :param str file_format: The format in which we want to export
    """
    logger = get_task_logger(__name__)
    logger.info("    + Exporting data to file")

    tmpdir = _get_tmp_directory_path()
    filename = format_filename(filename)
    if file_format == "ods":
        writer = OdsExporter()
    elif file_format == "xls":
        writer = XlsExporter()
    else:
        writer = CsvExporter()

    writer.headers = headers
    writer.set_datas(data)

    filepath = _get_tmp_filepath(tmpdir, filename, file_format)
    logger.debug(" + Writing file to %s" % filepath)
    with _get_open_file(filepath, file_format) as f_buf:
        writer.render(f_buf)
    result_filename = os.path.basename(filepath)

    logger.debug(f" -> The file {result_filename} been written")
    return result_filename


def amount_to_export_str(amount: int, precision: int = 5) -> str:
    """
    Convert amounts (as integers) in an exportable string
    """
    return round(integer_to_amount(amount, precision=precision, default=0), 2)


@celery_app.task(bind=True)
@cache_region("default_term", "estimations_details")
def export_estimations_details_to_file(self, job_id, task_ids, format="csv"):
    """
    Exporting details (each lines) of given estimations

    :param int job_id: The id of the job object
    :param list task_ids: List of estimations ids to query
    :param str format: The format in which we want to export
    """

    logger.info(
        f" + Exporting details for the following estimations's ids : {task_ids}"
    )

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    headers = (
        {"label": "Enseigne", "name": "company"},
        {"label": "N°Devis", "name": "estimation_number"},
        {"label": "Date", "name": "estimation_date"},
        {"label": "Client", "name": "customer"},
        {"label": "Description", "name": "description"},
        {"label": "Prix Unit. HT", "name": "unit_ht"},
        {"label": "Prix Unit. TTC", "name": "unit_ttc"},
        {"label": "Quantité", "name": "quantity"},
        {"label": "Unité", "name": "unity"},
        {"label": "TVA", "name": "tva"},
        {"label": "Total HT", "name": "total_ht"},
        {"label": "Total TVA", "name": "tva_amount"},
        {"label": "Total TTC", "name": "total_ttc"},
        {"label": "Compte produit", "name": "product"},
    )

    try:
        data = []
        query = (
            Task.query().filter(Task.id.in_(task_ids)).order_by(Task.status_date.desc())
        )
        for estimation in query:
            estimation_product = None
            logger.debug(f" + Collecting data for estimation {estimation.id}")
            for line in estimation.all_lines:
                logger.debug(f"    > Computing data for line {line.id}")
                line_product = line.product.name if line.product else ""
                row_data = {
                    "company": estimation.company.name,
                    "estimation_number": estimation.internal_number,
                    "estimation_date": format_date(estimation.date),
                    "customer": estimation.customer.label,
                    "description": strip_html_tags(line.description),
                    "unit_ht": amount_to_export_str(line.unit_ht()),
                    "unit_ttc": amount_to_export_str(line.unit_ttc()),
                    "quantity": line.quantity,
                    "unity": line.unity,
                    "tva": amount_to_export_str(max(line.get_tva(), 0), 4),
                    "total_ht": amount_to_export_str(line.total_ht()),
                    "tva_amount": amount_to_export_str(line.tva_amount()),
                    "total_ttc": amount_to_export_str(line.total()),
                    "product": line_product,
                }
                data.append(row_data)
                if estimation_product is None:
                    estimation_product = line_product
                elif line_product != estimation_product:
                    estimation_product = ""
            for discount in estimation.discounts:
                logger.debug(f"    > Computing data for discount {discount.id}")
                row_data = {
                    "company": estimation.company.name,
                    "estimation_number": estimation.internal_number,
                    "estimation_date": format_date(estimation.date),
                    "customer": estimation.customer.label,
                    "description": strip_html_tags(discount.description),
                    "unit_ht": amount_to_export_str(discount.total_ht() * -1),
                    "unit_ttc": amount_to_export_str(discount.total() * -1),
                    "quantity": 1,
                    "unity": "remise",
                    "tva": amount_to_export_str(max(discount.get_tva(), 0), 4),
                    "total_ht": amount_to_export_str(discount.total_ht() * -1),
                    "tva_amount": amount_to_export_str(discount.tva_amount() * -1),
                    "total_ttc": amount_to_export_str(discount.total() * -1),
                    "product": estimation_product,
                    "date": "",
                }
                data.append(row_data)
        logger.info(" + All estimations's details data where collected !")

        result_filename = _export_data_to_tabular_file(
            headers, data, filename="detail_devis_", file_format=format
        )
    except Exception:
        logger.exception("Error while generating export file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)
    else:
        utils.record_completed(FileGenerationJob, job_id, filename=result_filename)
    return ""


@celery_app.task(bind=True)
@cache_region("default_term", "invoices_details")
def export_invoices_details_to_file(self, job_id, task_ids, format="csv"):
    """
    Exporting details (each lines) of given invoices

    :param int job_id: The id of the job object
    :param list task_ids: List of invoices ids to query
    :param str format: The format in which we want to export
    """

    logger.info(f" + Exporting details for the following invoice's ids : {task_ids}")

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    headers = (
        {"label": "Enseigne", "name": "company"},
        {"label": "Facture", "name": "invoice_number"},
        {"label": "Date", "name": "invoice_date"},
        {"label": "Client", "name": "customer"},
        {"label": "Description", "name": "description"},
        {"label": "Prix Unit. HT", "name": "unit_ht"},
        {"label": "Prix Unit. TTC", "name": "unit_ttc"},
        {"label": "Quantité", "name": "quantity"},
        {"label": "Unité", "name": "unity"},
        {"label": "TVA", "name": "tva"},
        {"label": "Total HT", "name": "total_ht"},
        {"label": "Total TVA", "name": "tva_amount"},
        {"label": "Total TTC", "name": "total_ttc"},
        {"label": "Compte produit", "name": "product"},
        {"label": "Date d'exécution", "name": "date"},
    )

    try:
        data = []
        query = (
            Task.query().filter(Task.id.in_(task_ids)).order_by(Task.status_date.desc())
        )
        for invoice in query:
            invoice_product = None
            logger.debug(f" + Collecting data for invoice {invoice.id}")
            for line in invoice.all_lines:
                logger.debug(f"    > Computing data for line {line.id}")
                line_product = line.product.name if line.product else ""
                row_data = {
                    "company": invoice.company.name,
                    "invoice_number": invoice.official_number,
                    "invoice_date": format_date(invoice.date),
                    "customer": invoice.customer.label,
                    "description": strip_html_tags(line.description),
                    "unit_ht": amount_to_export_str(line.unit_ht()),
                    "unit_ttc": amount_to_export_str(line.unit_ttc()),
                    "quantity": line.quantity,
                    "unity": line.unity,
                    "tva": amount_to_export_str(max(line.get_tva(), 0), 4),
                    "total_ht": amount_to_export_str(line.total_ht()),
                    "tva_amount": amount_to_export_str(line.tva_amount()),
                    "total_ttc": amount_to_export_str(line.total()),
                    "product": line_product,
                    "date": format_date(line.date),
                }
                data.append(row_data)
                if invoice_product is None:
                    invoice_product = line_product
                elif line_product != invoice_product:
                    invoice_product = ""
            for discount in invoice.discounts:
                logger.debug(f"    > Computing data for discount {discount.id}")
                row_data = {
                    "company": invoice.company.name,
                    "invoice_number": invoice.official_number,
                    "invoice_date": format_date(invoice.date),
                    "customer": invoice.customer.label,
                    "description": strip_html_tags(discount.description),
                    "unit_ht": amount_to_export_str(discount.total_ht() * -1),
                    "unit_ttc": amount_to_export_str(discount.total() * -1),
                    "quantity": 1,
                    "unity": "remise",
                    "tva": amount_to_export_str(max(discount.get_tva(), 0), 4),
                    "total_ht": amount_to_export_str(discount.total_ht() * -1),
                    "tva_amount": amount_to_export_str(discount.tva_amount() * -1),
                    "total_ttc": amount_to_export_str(discount.total() * -1),
                    "product": invoice_product,
                    "date": "",
                }
                data.append(row_data)
        logger.info(" + All invoice's details data where collected !")

        result_filename = _export_data_to_tabular_file(
            headers, data, filename="detail_factures_", file_format=format
        )
    except Exception:
        logger.exception("Error while generating export file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)
    else:
        utils.record_completed(FileGenerationJob, job_id, filename=result_filename)
    return ""


@celery_app.task(bind=True)
@cache_region("default_term", "expenses")
def export_expenses_to_file(self, job_id, expense_ids, filename, format="csv"):
    """
    export expenses to file

    :param int job_id: The id of the job object
    :param list expense_ids: List of expenses ids to query
    :param str format: The format in which we want to export
    """

    logger.debug(
        f" + Exporting details for the following expenses's ids : {expense_ids}"
    )

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    headers = (
        {"label": "Numéro de pièce", "name": "official_number"},
        {"label": "Titre", "name": "title"},
        {"label": "Entrepreneur", "name": "user_name"},
        {"label": "Enseigne", "name": "company_name"},
        {"label": "Année", "name": "year"},
        {"label": "Mois", "name": "month"},
        {"label": "HT", "name": "ht"},
        {"label": "TVA", "name": "tva"},
        {"label": "TTC", "name": "ttc"},
        {"label": "Km", "name": "km"},
        {"label": "Justificatifs reçus et acceptés", "name": "justified"},
        {"label": "Paiements", "name": "paid"},
        {"label": "Montant restant dû", "name": "topay"},
    )

    expense_justified_status = {
        0: "non",
        1: "oui",
    }
    expense_paid_status = {
        "resulted": "intégral",
        "paid": "partiel",
        "waiting": "en attente",
    }

    try:
        data = []
        query = ExpenseSheet.query().filter(ExpenseSheet.id.in_(expense_ids))
        for expense in query:
            # compute topay column
            topay = 0
            if hasattr(expense, "topay") and not expense.paid_status == "resulted":
                topay = integer_to_amount(expense.topay())

            row_data = {
                "official_number": expense.official_number,
                "title": expense.title,
                "user_name": expense.user.label,
                "company_name": expense.company.name,
                "year": expense.year,
                "month": expense.month,
                "ht": integer_to_amount(expense.total_ht),
                "tva": integer_to_amount(expense.total_tva),
                "ttc": integer_to_amount(expense.total),
                "km": integer_to_amount(expense.total_km),
                "justified": expense_justified_status.get(
                    expense.justified, expense.justified
                ),
                "paid": expense_paid_status.get(
                    expense.paid_status, expense.paid_status
                ),
                "topay": topay,
            }
            data.append(row_data)
        result_filename = _export_data_to_tabular_file(
            headers, data, filename=filename, file_format=format
        )
    except Exception:
        logger.exception("Error while generating export file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)
    else:
        utils.record_completed(FileGenerationJob, job_id, filename=result_filename)
    return ""


@celery_app.task(bind=True)
def export_dataquery_to_file(
    self, job_id, dataquery_name, format="csv", start=None, end=None
):
    """
    Exporting the result of the given dataquery in a file with the given format

    :param int job_id: Id of the job object
    :param list dataquery_name: Name of the dataquery we want
    :param str format: The format in which we want to export
    :param str start: The start date of the query (if needed)
    :param str end: The end date of the query (if needed)
    """

    logger.debug(f" + Exporting results of dataquery : {dataquery_name}")

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    # Import and create the dataquery object from caerp
    try:
        pyramid_request = get_request()
        dataquery_object = pyramid_request.get_dataquery(dataquery_name)
        dataquery_object.set_dates(start, end)
    except Exception:
        logger.error(f"ABORT : Unable to instanciate dataquery '{dataquery_name}'")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)
        return ""

    # Generating dataquery headers and data
    headers = _transform_list_header_in_export_header(dataquery_object.headers())
    data = _transform_list_data_in_export_data(dataquery_object.data())
    logger.info(" + Query data collected !")

    # Exporting dataquery result in a file
    try:
        datequery_filename = f"{dataquery_object.name}_"
        if dataquery_object.start_date:
            datequery_filename += f"{dataquery_object.start_date.strftime('%Y-%m-%d')}_"
        if dataquery_object.end_date:
            datequery_filename += f"{dataquery_object.end_date.strftime('%Y-%m-%d')}_"
        result_filename = _export_data_to_tabular_file(
            headers, data, filename=datequery_filename, file_format=format
        )
    except Exception:
        logger.exception("Error while generating dataquery export file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)
    else:
        utils.record_completed(FileGenerationJob, job_id, filename=result_filename)
    return ""


@celery_app.task(bind=True)
def export_company_sales_catalog_to_json(
    self,
    job_id,
    company_id,
):
    """
    Exports a sales catalog from a company to a JSON file

    Sales catalog is a complex nested data structure, thus tabular data is not appropriate for it.
    """

    logger = get_task_logger(__name__)

    logger.info(f" + Exporting sales catalog for company : {company_id}")

    # Mark job started
    utils.start_job(self.request, FileGenerationJob, job_id)

    try:
        company = Company.get(company_id)
        filename = format_filename(f"{today():%Y-%m-%d} export-catalog-{company.name}")
        file_extension = "json"

        tmpdir = _get_tmp_directory_path()

        filepath = _get_tmp_filepath(tmpdir, filename, file_extension)

        data = serialize_catalog(company)

        logger.debug(" + Writing file to %s" % filepath)

        with _get_open_file(filepath, file_extension) as f_buf:
            f_buf.write(json.dumps(data, indent=2))

        result_filename = os.path.basename(filepath)

    except Exception:
        logger.exception("Error while generating file")
        errors = [GENERATION_ERROR_MESSAGE % job_id]
        utils.record_failure(FileGenerationJob, job_id, errors)

    else:
        logger.debug(f" -> The file {result_filename} been written")
        utils.record_completed(FileGenerationJob, job_id, filename=result_filename)

        return result_filename



import json
import os

import transaction
from celery.utils.log import get_task_logger
from pyramid_celery import celery_app

from caerp.import_.sale_product import deserialize_catalog
from caerp.models.company import Company
from caerp_celery.models import CsvImportJob
from caerp_celery.tasks import utils


IMPORT_ERROR_MESSAGE = (
    "Une erreur inconnue a été rencontrée à l'import de votre fichier, "
    "veuillez contacter votre administrateur en lui "
    "fournissant l'identifiant suivant : %s"
)


@celery_app.task(bind=True)
def import_json_company_sales_catalog(
    self,
    job_id,
    company_id,
    filepath,
):
    from caerp_base.models.base import DBSESSION

    logger = get_task_logger(__name__)

    logger.info(f" + Importing sales catalog for company : {company_id}")
    try:
        # Mark job started
        utils.start_job(self.request, CsvImportJob, job_id)
        transaction.begin()
        dbsession = DBSESSION()

        company = Company.get(company_id)

        with open(filepath) as f:
            objs, warnings = deserialize_catalog(company, json.load(f))

            dbsession.add_all(objs)
            dbsession.flush()
            msg = (
                f"{len(objs)} produits importés dans "
                f"le catalogue de l'enseigne f{company.name}"
            )
            transaction.commit()

    except Exception:
        transaction.abort()
        logger.exception(f"Error while importing catalog for Company {company_id}")
        errors = [IMPORT_ERROR_MESSAGE % job_id]
        utils.record_failure(CsvImportJob, job_id, errors)
    else:
        utils.record_completed(
            CsvImportJob,
            job_id,
            error_messages=warnings,
            messages=[msg],
        )

    finally:
        if filepath:
            try:
                logger.info(f"Deleting file {filepath}")
                os.remove(filepath)
            except Exception as e:
                logger.error(f"Unable to delete file")

    return ""



import datetime
from sqlalchemy import delete
from caerp_celery.transactional_task import task_tm
from caerp.models.user.login import UserConnections
from caerp_celery.conf import get_request


@task_tm
def clean_old_user_connections_task():
    now = datetime.datetime.now()
    purge_datetime = now - datetime.timedelta(days=365)
    query = delete(UserConnections).where(
        UserConnections.month_last_connection < purge_datetime,
    )
    request = get_request()
    request.dbsession.execute(query)



# -*- coding: utf-8 -*-
"""
Custom exceptions
"""


class BaseMessageException(Exception):
    def __init__(self, message=""):
        self.message = message


class FileNameException(BaseMessageException):
    pass


class MissingMandatoryArgument(BaseMessageException):
    """
    Raised when a mandatory argument is missing
    """

    pass


class InstanceNotFound(BaseMessageException):
    """
    Raised when no instance could be found
    """

    pass


class MultipleInstanceFound(BaseMessageException):
    """
    Raised when no instance could be found
    """

    pass



# -*- coding: utf-8 -*-
"""
All asynchronous tasks runned through CAErp are stored here
Tasks are handled by a celery service
Redis is used as the central bus
"""
import transaction
from pyramid_celery import celery_app

from caerp_base.exception import (
    UndeliveredMail,
    MailAlreadySent,
)
from caerp_celery.mail import send_salary_sheet
from caerp_celery.conf import get_request
from caerp_celery.tasks import utils
from caerp_celery.models import (
    MailingJob,
)


logger = utils.get_logger(__name__)


@celery_app.task(bind=True)
def async_mail_salarysheets(self, job_id, mails, force):
    """
    Asynchronously sent a bunch of emails with attached salarysheets

    :param int job_id: The id of the MailSendJob
    :param mails: a list of dict compound of
        {
            'id': company_id,
            'attachment': attachment filename,
            'attachment_path': attachment filepath,
            'message': The mail message,
            'subject': The mail subject,
            'company_id': The id of the company,
            'email': The email to send it to,
        }
    :param force: Should we force the mail sending
    """
    logger.info("We are launching an asynchronous mail sending operation")
    logger.info("  The job id : %s" % job_id)

    from caerp_base.models.base import DBSESSION

    # Mark job started
    utils.start_job(self.request, MailingJob, job_id)

    mail_count = 0
    error_count = 0
    error_messages = []
    request = get_request()
    session = DBSESSION()
    for mail_datas in mails:
        transaction.begin()
        # since we send a mail out of the transaction process, we need to
        # commit each mail_history instance to avoid sending and not storing
        # the history
        try:
            company_id = mail_datas["company_id"]
            email = mail_datas["email"]

            if email is None:
                logger.error("no mail found for company {0}".format(company_id))
                continue
            else:
                message = mail_datas["message"]
                subject = mail_datas["subject"]
                logger.info("  The mail subject : %s" % subject)
                logger.info("  The mail message : %s" % message)

                mail_history = send_salary_sheet(
                    request,
                    email,
                    company_id,
                    mail_datas["attachment"],
                    mail_datas["attachment_path"],
                    force=force,
                    message=message,
                    subject=subject,
                )
                # Stores the history of this sent email
                session.add(mail_history)
            transaction.commit()

        except MailAlreadySent:
            transaction.abort()
            error_count += 1
            msg = "Ce fichier a déjà été envoyé {0}".format(mail_datas["attachment"])
            error_messages.append(msg)
            logger.exception("Mail already delivered")
            logger.error("* Part of the Task FAILED")
            continue

        except UndeliveredMail:
            transaction.abort()
            error_count += 1
            msg = "Impossible de délivrer de mail à l'entreprise {0} \
(mail : {1})".format(
                company_id, email
            )
            error_messages.append(msg)
            logger.exception("Unable to deliver an e-mail")
            logger.error("* Part of the Task FAILED")
            continue

        except Exception as e:
            transaction.abort()
            error_count += 1
            logger.exception("The subtransaction has been aborted")
            logger.error("* Part of the task FAILED !!!")
            error_messages.append("{0}".format(e))

        else:
            mail_count += 1
            logger.info("The transaction has been commited")
            logger.info("* Part of the Task SUCCEEDED !!!")

    messages = ["{0} mails ont été envoyés".format(mail_count)]
    messages.append("{0} mails n'ont pas pu être envoyés".format(error_count))

    logger.info("-> Task finished")
    if error_count > 0:
        utils.record_failure(
            MailingJob,
            job_id,
            error_messages=error_messages,
            messages=messages,
        )
    else:
        utils.record_completed(
            MailingJob,
            job_id,
            messages=messages,
        )



import functools

from sqlalchemy import desc, func

from caerp.compute.math_utils import (
    integer_to_amount,
    translate_integer_precision,
)
from caerp.consts.permissions import PERMISSIONS
from caerp.models.accounting.bookeeping import CustomInvoiceBookEntryModule
from caerp.models.accounting.operations import AccountingOperation
from caerp.models.supply import SupplierInvoice, InternalSupplierInvoice
from caerp.models.tva import Tva
from caerp.models.task import Task
from caerp.models.third_party.customer import Customer
from caerp.models.third_party.supplier import Supplier
from caerp.models.user.userdatas import UserDatas
from caerp.utils.renderer import configure_export
from caerp.utils.strings import format_supplier_invoice_status


def _add_userdatas_custom_headers(writer, query):
    """
    Specific to userdatas exports

    Add custom headers that are not added through automation

    Add headers for code_compta
    """
    from caerp_base.models.base import DBSESSION
    from caerp.models.user.user import COMPANY_EMPLOYEE

    # Compte analytique
    query = DBSESSION().query(func.count(COMPANY_EMPLOYEE.c.company_id).label("nb"))
    query = query.group_by(COMPANY_EMPLOYEE.c.account_id)
    code_compta_count = query.order_by(desc("nb")).first()
    if code_compta_count:
        code_compta_count = code_compta_count[0]
        for index in range(0, code_compta_count):
            new_header = {
                "label": "Compte_analytique {0}".format(index + 1),
                "name": "code_compta_{0}".format(index + 1),
            }
            writer.add_extra_header(new_header)

    return writer


def _add_userdatas_code_compta(writer, userdatas):
    """
    Add code compta to exports (specific for userdatas exports)

    :param obj writer: The tabbed file writer
    :param obj userdatas: The UserDatas instance we manage
    """
    user_account = userdatas.user
    if user_account:
        datas = []
        for company in user_account.companies:
            datas.append(company.code_compta)
        writer.add_extra_datas(datas)
    return writer


def _add_invoice_custom_headers(writer, query):
    """
    Invoice export headers

    * Entreprise ;
    • Client ;
    • Date ;
    • Numéro de factures ;
    • Objet ;
    • Lieu des travaux ;
    • HT (par taux de tva) ;
    • TVA (par taux de tva) ;
    • TTC.
    """
    return _add_task_custom_headers(writer, query, topay=True)


def _add_estimation_custom_headers(writer, query):
    """
    Estimation export headers

    * Entreprise ;
    • Client ;
    • Date ;
    • Objet ;
    • Lieu des travaux ;
    • HT (par taux de tva) ;
    • TVA (par taux de tva) ;
    • TTC.
    """
    return _add_task_custom_headers(writer, query, topay=False)


def _add_task_custom_headers(writer, query, topay):
    writer.add_extra_header({"label": "Montant HT", "name": "ht"})
    writer.add_extra_header({"label": "Montant TTC", "name": "ttc"})
    if topay:
        writer.add_extra_header({"label": "Montant restant dû", "name": "topay"})
    for tva in Tva.query():
        writer.add_extra_header(
            {
                "label": "HT {tva.name}".format(tva=tva),
                "name": "HT {tva.value}".format(tva=tva),
            }
        )
        writer.add_extra_header({"label": tva.name, "name": tva.value})

    writer.add_extra_header({"label": "Assurance", "name": "insurance_label"})
    writer.add_extra_header({"label": "Taux d'assurance", "name": "insurance_rate"})
    writer.add_extra_header(
        {"label": "Montant de l'assurance", "name": "insurance_amount"}
    )
    return writer


def _add_invoice_datas(writer, invoice):
    _has_insurance_module = False
    if CustomInvoiceBookEntryModule.get_by_name("insurance", invoice.prefix):
        _has_insurance_module = True

    ht_values = invoice.tva_ht_parts()
    tva_values = invoice.get_tvas()
    datas = []
    datas.append(integer_to_amount(invoice.ht, precision=5))
    datas.append(integer_to_amount(invoice.ttc, precision=5))
    if hasattr(invoice, "topay") and not invoice.paid_status == "resulted":
        datas.append(integer_to_amount(invoice.topay(), precision=5))
    else:
        datas.append(0)
    for tva in Tva.query():
        datas.append(integer_to_amount(ht_values.get(tva.value, 0), precision=5))
        datas.append(integer_to_amount(tva_values.get(tva.value, 0), precision=5))

    if invoice.insurance:
        datas.append(invoice.insurance.label)
    else:
        if not _has_insurance_module:
            datas.append("")
        else:
            level = invoice.get_rate_level("insurance")
            if level == "cae":
                datas.append("Taux de la CAE (module assurance)")
            elif level == "company":
                datas.append("Taux d'assurance de l'enseigne")
            else:
                datas.append("")

    if _has_insurance_module:
        rate = invoice.get_rate("insurance")
        if not rate:
            rate = 0
    else:
        if invoice.insurance:
            rate = invoice.insurance.rate
        else:
            rate = 0
    datas.append("{} %".format(rate))

    if _has_insurance_module:
        insurance = invoice.total_insurance(invoice.ht)
        rounded = translate_integer_precision(insurance)
    else:
        if invoice.insurance:
            insurance = invoice.total_insurance(invoice.ht)
            rounded = translate_integer_precision(insurance)
        else:
            rounded = 0
    datas.append(integer_to_amount(rounded, precision=2))
    writer.add_extra_datas(datas)
    return writer


def _add_estimation_datas(writer, invoice):
    _has_insurance_module = False
    if CustomInvoiceBookEntryModule.get_by_name("insurance", invoice.prefix):
        _has_insurance_module = True

    ht_values = invoice.tva_ht_parts()
    tva_values = invoice.get_tvas()
    datas = []
    datas.append(integer_to_amount(invoice.ht, precision=5))
    datas.append(integer_to_amount(invoice.ttc, precision=5))
    for tva in Tva.query():
        datas.append(integer_to_amount(ht_values.get(tva.value, 0), precision=5))
        datas.append(integer_to_amount(tva_values.get(tva.value, 0), precision=5))

    if invoice.insurance:
        datas.append(invoice.insurance.label)
    else:
        if not _has_insurance_module:
            datas.append("")
        else:
            level = invoice.get_rate_level("insurance")
            if level == "cae":
                datas.append("Taux de la CAE (module assurance)")
            elif level == "company":
                datas.append("Taux d'assurance de l'enseigne")
            else:
                datas.append("")

    if _has_insurance_module:
        rate = invoice.get_rate("insurance")
        if not rate:
            rate = 0
    else:
        if invoice.insurance:
            rate = invoice.insurance.rate
        else:
            rate = 0
    datas.append("{} %".format(rate))

    if _has_insurance_module:
        insurance = invoice.total_insurance(invoice.ht)
        rounded = translate_integer_precision(insurance)
    else:
        if invoice.insurance:
            insurance = invoice.total_insurance(invoice.ht)
            rounded = translate_integer_precision(insurance)
        else:
            rounded = 0
    datas.append(integer_to_amount(rounded, precision=2))
    writer.add_extra_datas(datas)
    return writer


def _add_supplier_invoice_custom_headers(writer, supplier_invoices):
    writer.add_extra_header({"label": "Montant HT", "name": "ht"})
    writer.add_extra_header({"label": "Montant TVA", "name": "tva"})
    writer.add_extra_header({"label": "Montant TTC", "name": "ttc"})
    writer.add_extra_header({"label": "Type", "name": "typ"})
    writer.add_extra_header({"label": "Statut", "name": "status"})
    return writer


def _add_supplier_invoice_data(writer, supplier_invoice):
    """Add Supplier invoice related information to the export"""
    typ = "Externe"
    if isinstance(supplier_invoice, InternalSupplierInvoice):
        typ = "Interne"

    status = format_supplier_invoice_status(supplier_invoice, full=False)

    writer.add_extra_datas(
        [
            integer_to_amount(supplier_invoice.total_ht, precision=2),
            integer_to_amount(supplier_invoice.total_tva, precision=2),
            integer_to_amount(supplier_invoice.total, precision=2),
            typ,
            status,
        ]
    )
    return writer


def _import_userdatas_add_related_user(action, dbsession, model, updated, args):
    """
    Creat a User instance when importing a new UserDatas instance

    :param str action: An import action (override, insert, update, only_update,
    only_override)
    :param obj dbsession: The database transaction session (DBSESSION)
    :param obj model: The newly inserted model
    :param bool updated: Is it an update of an existing model ?
    :param dict args: The importation arguments
    """
    if action == "insert":
        model.gen_related_user_instance()
    return model


def _import_third_party_set_label(action, dbsession, model, updated, args):
    """
    Ensure the currently managed model has a label set
    """
    if not model.label:
        model.label = model._get_label()
        model = dbsession.merge(model)
    if not model.name:
        model.name = model.label
        model = dbsession.merge(model)
    return model


def _import_third_party_set_type_(type_, action, dbsession, model, updated, args):
    """
    Set the polymorphic key type_ on the model
    """
    if not model.type_:
        model.type_ = type_
        model = dbsession.merge(model)
    return model


def includeme(config):
    configure_export()
    config.register_import_model(
        key="userdatas",
        model=UserDatas,
        label="Données de gestion sociale",
        permission=PERMISSIONS["global.view_userdata_details"],
        excludes=(
            "name",
            "created_at",
            "updated_at",
            "type_",
            "_acl",
            "parent_id",
            "parent",
        ),
        callbacks=[_import_userdatas_add_related_user],
    )
    config.register_import_model(
        key="customers",
        model=Customer,
        label="Clients",
        permission=PERMISSIONS["context.add_customer"],
        excludes=(
            "created_at",
            "updated_at",
            "company_id",
            "company",
            "type_",
        ),
        callbacks=[
            _import_third_party_set_label,
            functools.partial(_import_third_party_set_type_, "customer"),
        ],
    )
    config.register_import_model(
        key="suppliers",
        model=Supplier,
        label="Fournisseurs",
        permission=PERMISSIONS["context.add_supplier"],
        excludes=(
            "created_at",
            "updated_at",
            "company_id",
            "company",
            "type_",
        ),
        callbacks=[
            _import_third_party_set_label,
            functools.partial(_import_third_party_set_type_, "customer"),
        ],
    )
    config.register_export_model(
        key="userdatas",
        model=UserDatas,
        options={
            "hook_init": _add_userdatas_custom_headers,
            "hook_add_row": _add_userdatas_code_compta,
            "foreign_key_name": "userdatas_id",
            "excludes": (
                "name",
                "type_",
                "_acl",
                "parent_id",
                "parent",
            ),
        },
    )
    config.register_export_model(
        key="invoices",
        model=Task,
        options={
            "hook_init": _add_invoice_custom_headers,
            "hook_add_row": _add_invoice_datas,
            "foreign_key_name": "task_id",
            "excludes": ("name", "created_at", "updated_at", "type_"),
            "order": (
                "company",
                "customer",
                "date",
                "official_number",
                "description",
                "workplace",
            ),
        },
    )
    config.register_export_model(
        key="supplier_invoices",
        model=SupplierInvoice,
        options={
            "hook_init": _add_supplier_invoice_custom_headers,
            "hook_add_row": _add_supplier_invoice_data,
            "foreign_key_name": "supplier_invoice_id",
            "excludes": (
                "name",
                "created_at",
                "updated_at",
                "type_",
                "supplier_orders",
                "lines",
                "exports",
                "exported",
            ),
            "order": (
                "company",
                "supplier",
                "date",
                "official_number",
                "remote_invoice_number",
                "exported",
                "cae_percentage",
                "payer",
            ),
        },
    )
    config.register_export_model(
        key="accounting_operations",
        model=AccountingOperation,
        options={
            "foreign_key_name": "id",
            "order": (
                "date",
                "general_account",
                "label",
                "debit",
                "credit",
                "balance",
                "analytical_account",
                "company",
            ),
        },
    )
    config.register_export_model(
        key="estimations",
        model=Task,
        options={
            "hook_init": _add_estimation_custom_headers,
            "hook_add_row": _add_estimation_datas,
            "foreign_key_name": "task_id",
            "excludes": ("name", "created_at", "updated_at", "type_"),
            "order": (
                "company",
                "customer",
                "date",
                "description",
                "workplace",
            ),
        },
    )



# -*- coding: utf-8 -*-
import logging
from typing import List
from pyramid_mailer.message import (
    Attachment,
)

from caerp_base.exception import (
    MailAlreadySent,
    UndeliveredMail,
)
from caerp_base.mail import send_mail
from caerp_celery.models import (
    store_sent_mail,
    check_if_mail_sent,
)
from caerp.models.third_party import Customer
from caerp.models.user import User, UserDatas


logger = logging.getLogger(__name__)


SALARYSHEET_MAIL_MESSAGE = """Bonjour,
Vous trouverez ci-joint votre bulletin de salaire.
"""

SALARYSHEET_MAIL_SUBJECT = "Votre bulletin de salaire"


def send_salary_sheet(
    request,
    email,
    company_id,
    filename,
    filepath,
    force=False,
    message=None,
    subject=None,
):
    """
    Send a salarysheet to the given company's e-mail

    :param obj request: A pyramid request object
    :param str company_mail: The mail to send it to
    :param int company_id: The id of the associated company
    :param str filepath: The path to the filename
    :param bool force: Whether to force sending this file again
    :param str message: The mail message
    :param str subject: The mail subject
    :returns: A MailHistory instance
    :TypeError UndeliveredMail: When the company has no mail
    :TypeError MailAlreadySent: if the file has
        already been sent and no force option was passed
    """
    filebuf = open(filepath, "rb")
    filedatas = filebuf.read()

    if not force and check_if_mail_sent(filedatas, company_id):
        logger.warning("Mail already sent : mail already sent")
        raise MailAlreadySent("Mail already sent")

    filebuf.seek(0)

    if email is None:
        logger.warning(
            "Undelivered email : no mail provided for company {0}".format(company_id)
        )
        raise UndeliveredMail("no mail provided for company {0}".format(company_id))
    else:
        logger.info("Sending the file %s" % filepath)
        logger.info("Sending it to %s" % email)
        attachment = Attachment(filename, "application/pdf", filebuf)

        subject = subject or SALARYSHEET_MAIL_SUBJECT
        message = message or SALARYSHEET_MAIL_MESSAGE

        send_mail(
            request,
            email,
            message,
            subject,
            attachment,
        )
        return store_sent_mail(filepath, filedatas, company_id)


INTERNAL_ORDER_CUSTOMER_MAIL_OBJECT = """Nouvelle commande founisseur générée \
dans votre espace"""

INTERNAL_ORDER_CUSTOMER_MAIL_TMPL = """
Bonjour {customer},

L'enseigne {supplier} vous a transmis un devis 'interne'.

Une commande fournisseur contenant le devis est accessible dans votre espace
dans la section "Commande fournisseur".
"""

INTERNAL_ORDER_SUPPLIER_MAIL_OBJECT = "Votre devis a été transmis à votre \
client"
INTERNAL_ORDER_SUPPLIER_MAIL_TMPL = """
Bonjour {supplier},

Votre devis 'interne' a été transmis à l'enseigne {customer}.

Une commande fournisseur contenant le devis est accessible dans son espace dans
la section "Commande fournisseur".
"""


def send_customer_new_order_mail(request, order):
    """
    Send an email to an internal customer
    """
    customer = order.company
    supplier = order.supplier
    message = INTERNAL_ORDER_CUSTOMER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if customer.email:
        recipients = [customer.email]
        send_mail(request, recipients, message, INTERNAL_ORDER_CUSTOMER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(customer.id))


def send_supplier_new_order_mail(request, order):
    """
    Send an email to an internal supplier
    """
    customer = order.company
    supplier = order.supplier
    message = INTERNAL_ORDER_SUPPLIER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if supplier:
        recipients = [supplier.email]
        send_mail(request, recipients, message, INTERNAL_ORDER_SUPPLIER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(supplier.id))


INTERNAL_INVOICE_CUSTOMER_MAIL_OBJECT = """Nouvelle facture founisseur \
générée dans votre espace"""

INTERNAL_INVOICE_CUSTOMER_MAIL_TMPL = """
Bonjour {customer},

L'enseigne {supplier} vous a transmis une facture 'interne'.

Une facture fournisseur contenant la facture est accessible dans votre espace
dans la section "Facture fournisseur".
"""

INTERNAL_INVOICE_SUPPLIER_MAIL_OBJECT = "Votre facture a été transmis à votre \
client"
INTERNAL_INVOICE_SUPPLIER_MAIL_TMPL = """
Bonjour {supplier},

Votre facture 'interne' a été transmis à l'enseigne {customer}.

Une facture fournisseur contenant la facture est accessible dans son espace
dans la section "Facture fournisseur".
"""


def send_customer_new_invoice_mail(request, supplier_invoice):
    """
    Send an email to an internal customer
    """
    customer = supplier_invoice.company
    supplier = supplier_invoice.supplier
    message = INTERNAL_INVOICE_CUSTOMER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if customer.email:
        recipients = [customer.email]
        send_mail(request, recipients, message, INTERNAL_INVOICE_CUSTOMER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(customer.id))


def send_supplier_new_invoice_mail(request, supplier_invoice):
    """
    Send an email to an internal supplier
    """
    customer = supplier_invoice.company
    supplier = supplier_invoice.supplier
    message = INTERNAL_INVOICE_SUPPLIER_MAIL_TMPL.format(
        customer=customer.name, supplier=supplier.label
    )
    if supplier:
        recipients = [supplier.email]
        send_mail(request, recipients, message, INTERNAL_INVOICE_SUPPLIER_MAIL_OBJECT)
    else:
        logger.error("Company {} has no email set".format(supplier.id))


SUPPORT_MESSAGE = (
    "Compte tenu du nombre important de données à anonymiser, "
    "vous pouvez faire appel à votre support pour vous assister."
)
USERDATA_EMAIL_SUBJECT = "[RGPD] : Il y a des fiches de gestion sociale à nettoyer"
USERDATA_EMAIL_BODY = """Bonjour,

Vous recevez cet e-mail en tant que responsable RGPD.
{additional_message}

Les {number} fiches de gestion sociale suivantes  
ne semblent pas avoir été utilisées depuis plus de {days} jours. 
Elles devraient être anonymisées.

{additional_message}

{email_body}
"""


def _get_userdata_link(request, userdata_entry):
    from caerp.views.userdatas.routes import USER_USERDATAS_URL

    settings = request.registry.settings
    instance_name = settings.get("caerp.instance_name")
    view_path = USER_USERDATAS_URL.format(id=userdata_entry.user_id)
    return f"https://{instance_name}/{view_path}"


def send_rgpd_userdata_notification(
    request, recipients, userdata_entries: List[UserDatas], days: int
):
    """
    Send an email to the RGPD team notifying about userdata entries that should be
    cleaned up.

    :param recipients: List of email addresses to send the notification to.
    :param userdata_entries: List of Userdata instances that should be cleaned up.
    :param days: Number of days after which userdata entries should be considered old.
    """
    subject = USERDATA_EMAIL_SUBJECT.format(number=len(userdata_entries))
    links = ""
    if len(userdata_entries) > 10:
        additional_message = SUPPORT_MESSAGE
    else:
        additional_message = ""
    for entry in userdata_entries:
        entry_url = _get_userdata_link(request, entry)
        links += (
            f"{entry.coordonnees_lastname} - {entry.coordonnees_firstname} - "
            f"{entry_url}\n"
        )
    body = USERDATA_EMAIL_BODY.format(
        number=len(userdata_entries),
        days=days,
        email_body=links,
        additional_message=additional_message,
    )
    send_mail(request, recipients, body, subject)


USER_ACCOUNT_EMAIL_SUBJECT = "[RGPD] : Il y a des comptes à désactiver"
USER_ACCOUNT_EMAIL_BODY = """Bonjour,

Vous recevez cet e-mail en tant que responsable RGPD.

Les {number} comptes suivants sont toujours actifs mais n'ont pas été utilisés 
depuis plus de {days} jours.
Ils devraient être désactivés.

{additional_message}

{email_body}
"""


def _get_user_account_link(request, user: User):
    from caerp.views.user.routes import USER_LOGIN_URL

    settings = request.registry.settings
    instance_name = settings.get("caerp.instance_name")
    view_path = USER_LOGIN_URL.format(id=user.id)
    return f"https://{instance_name}/{view_path}"


def send_unused_login_notification(
    request, recipients, user_entries: List[User], days: int
):
    """
    Send an email to the RGPD team notifying about unused active accounts

    :param recipients: List of email addresses to send the notification to.
    :param userdata_entries: List of Userdata instances that should be cleaned up.
    :param days: Number of days after which userdata entries should be considered old.
    """
    subject = USER_ACCOUNT_EMAIL_SUBJECT.format(number=len(user_entries))
    links = ""
    if len(user_entries) > 10:
        additional_message = SUPPORT_MESSAGE
    else:
        additional_message = ""
    for entry in user_entries:
        entry_url = _get_user_account_link(request, entry)
        links += f"{entry.lastname} - {entry.firstname} - " f"{entry_url}\n"
    body = USER_ACCOUNT_EMAIL_BODY.format(
        number=len(user_entries),
        days=days,
        email_body=links,
        additional_message=additional_message,
    )
    send_mail(request, recipients, body, subject)


CUSTOMER_EMAIL_SUBJECT = "[RGPD] : Il y a des fiches clients à nettoyer"
CUSTOMER_EMAIL_BODY = """Bonjour,

Vous recevez cet e-mail en tant que responsable RGPD.

Les {number} comptes client (particulier) suivants sont toujours actifs mais n'ont pas été utilisés depuis plus de {days} jours. Ils devraient être anonymisés.

{additional_message}

{email_body}
"""


def _get_customer_link(request, customer: Customer):
    from caerp.views.third_party.customer.routes import CUSTOMER_ITEM_ROUTE

    settings = request.registry.settings
    instance_name = settings.get("caerp.instance_name")
    view_path = CUSTOMER_ITEM_ROUTE.format(id=customer.id)
    return f"https://{instance_name}/{view_path}"


def send_rgpd_customer_notification(
    request, recipients, customers: List[Customer], days: int
):
    """
    Send an email to the RGPD team notifying about unused individual customer accounts

    :param recipients: List of email addresses to send the notification to.
    :param userdata_entries: List of Userdata instances that should be cleaned up.
    :param days: Number of days after which userdata entries should be considered old.
    """
    subject = CUSTOMER_EMAIL_SUBJECT.format(number=len(customers))
    if len(customers) > 10:
        additional_message = SUPPORT_MESSAGE
    else:
        additional_message = ""
    links = ""
    for entry in customers:
        entry_url = _get_customer_link(request, entry)
        links += f"{entry.lastname} - {entry.firstname} - " f"{entry_url}\n"
    body = CUSTOMER_EMAIL_BODY.format(
        number=len(customers), days=days, email_body=links, 
        additional_message=additional_message,
    )
    send_mail(request, recipients, body, subject)



from caerp_celery.tasks import utils
from caerp_celery.transactional_task import task_tm
from caerp_celery.locks import release_lock

logger = utils.get_logger(__name__)


@task_tm
def release_lock_after_commit(lockname):
    """Delay the release of a lock after a commited transaction (used for numbering)"""
    logger.debug("Releasing lock {}".format(lockname))
    release_lock(lockname)



# -*- coding: utf-8 -*-
"""
Accounting operations parsing

Parses :

    csv files
    slk files

- Collect the administration mail address
- Inspect the pool for waiting files
- Find the parser and the filetype
- Parse
    - Collect file informations
    - Ensure we have csv datas
    - Insert lines in database

"""
import os
import datetime
import transaction

from dateutil.relativedelta import relativedelta
from pathlib import Path
from pyramid_celery import celery_app
from sqlalchemy import distinct
from typing import List, Optional, Tuple

from caerp_base.mail import send_mail
from caerp_base.models.base import DBSESSION
from caerp_celery.conf import (
    get_setting,
    get_recipients_addresses,
    get_request,
)
from caerp_celery.exception import FileNameException
from caerp_celery.interfaces import (
    IAccountingOperationProducer,
    IAccountingFileParser,
)
from caerp_celery.parsers import BaseParser, BaseProducer, UploadMetaData
from caerp_celery.tasks import utils
from caerp_celery.tasks.accounting_measure_compute import (
    compile_measures,
    get_exercice_dates_from_date,
)
from caerp.models.company import Company
from caerp.models.accounting.operations import (
    AccountingOperationUpload,
    AccountingOperation,
)
from caerp.utils.accounting import get_cae_accounting_software


logger = utils.get_logger(__name__)
FILENAME_ERROR = (
    "Le fichier ne respecte pas la nomenclature de nom attendue (ex : "
    "2017_12_resultat.slk)",
)

MAIL_ERROR_SUBJECT = "[ERREUR] Traitement de votre document \
{filename}"

MAIL_ERROR_BODY = """Une erreur est survenue lors du traitement du
fichier {filename}:

    {error}
"""
MAIL_UNKNOWN_ERROR_BODY = """Une erreur inconnue est survenue lors du
traitement du fichier {filename}:

    {error}

Veuillez contacter votre administrateur
"""
MAIL_SUCCESS_SUBJECT = """[Succès] Traitement de votre document {0}"""
MAIL_SUCCESS_BODY = """Le fichier {0} a été traité avec succès.
Écritures générées : {1}
Écritures n'ayant pas pu être associées à une entreprise existante : {2}

Les indicateurs ont été générés depuis ces écritures.
"""


def _get_base_path() -> Path:
    """
    Retreive the base working path as configured in the ini file
    """
    return Path(get_setting("caerp.parsing_pool_parent", mandatory=True))


def _get_path(directory) -> Path:
    """
    Return the abs path for the given directory

    :param str directory: The directory name pool/error/processed
    :rtype: str
    """
    return _get_base_path().joinpath(directory)


def _get_file_path_from_pool(pool_path: Path) -> Optional[Path]:
    """
    Handle file remaining in the pool

    :param pool_path: The pool path to look into
    :returns: The Path of the first file we find in the rep
    :rtype: str
    """
    result = None
    if pool_path.is_dir():
        for path in pool_path.iterdir():
            if path.is_file():
                result = path
                break
    return result


def _mv_file(file_path: Path, queue: str = "processed"):
    """
    Move the file to the processed directory
    """
    if file_path.is_file():
        new_dir = _get_path(queue)
        new_file_path = os.path.join(new_dir, file_path.name)
        os.system('mv "{0}" "{1}"'.format(file_path, new_file_path))
        logger.info(
            "The file {0} has been moved to the {1} directory".format(
                file_path,
                new_dir,
            )
        )
        return new_file_path
    else:
        raise Exception("File is missing {}".format(file_path))


def _clean_old_operations(old_ids, date_ref=None):
    """
    Clean old AccountingOperation entries

    :param list old_ids: The ids of items to remove
    :param date date_ref: The date from where we remove the items
    """
    op = AccountingOperation.__table__.delete().where(
        AccountingOperation.id.in_(old_ids)
    )
    if date_ref:
        op = op.where(AccountingOperation.date >= date_ref)
    op.execute()


class AccountingDataHandler:
    """
    Accounting datas parser : parses a General Ledger file
    """

    def __init__(self, parser: BaseParser, producer: BaseProducer, force: bool = False):
        self.parser = parser
        self.producer = producer
        self.force = force
        self.company_id_cache = {}

    def _load_company_id_cache(self):
        """
        Load company ids in a cache to avoid looking for them with each operation we
        treat
        """
        query = (
            DBSESSION()
            .query(distinct(Company.code_compta))
            .filter(Company.code_compta != None)  # noqa: E711
        )
        for (code_compta,) in query:
            company_id = Company.get_id_by_analytical_account(code_compta)
            self.company_id_cache[code_compta] = company_id

    def _find_company_id(self, analytical_account: str) -> Optional[int]:
        """
        Find a company object starting from its analytical_account

        :param str analytical_account: The account
        :returns: The company's id
        """
        return self.company_id_cache.get(analytical_account)

    def _build_operation_upload_object(self) -> AccountingOperationUpload:
        """
        Build an AccountingOperationUpload instance with current file datas

        :returns: An AccountingOperationUpload for the current parsed file
        """
        upload_data: UploadMetaData = self.parser.metadata()
        return AccountingOperationUpload(**upload_data.__dict__)

    def _get_existing_operation_ids(self) -> List[int]:
        """
        Return ids of the operations already stored in database
        """
        return [entry[0] for entry in DBSESSION().query(AccountingOperation.id)]

    def _get_upload_infos_from_operations_stream(
        self, exercice_start_date
    ) -> tuple[Optional[datetime.date], Optional[datetime.date], bool]:
        """
        Return needed informations from streamed operations

        :returns date first_date: date of the first operation in chronological order
        :returns bool is_previous_closure_done: if we can consider that previous fiscal
        year has been closed or not, based on existance of 6 or 7 operations before
        actual fiscal year start
        """
        first_date = None
        is_previous_closure_done = True
        last_date = None
        today = datetime.date.today()
        for operation in self.producer.stream_operations():
            if first_date is None or operation.date < first_date:
                first_date = operation.date

            if last_date is None or operation.date > last_date:
                last_date = min(operation.date, today)
            if (
                operation.general_account[:1] in ("6", "7")
                and operation.date < exercice_start_date
            ):
                is_previous_closure_done = False
        return first_date, last_date, is_previous_closure_done

    def run(self) -> Tuple[int, int, List[int], Optional[datetime.date]]:
        self._load_company_id_cache()
        upload = self._build_operation_upload_object()
        old_ids = self._get_existing_operation_ids()
        (exercice_start, exercice_end) = get_exercice_dates_from_date(upload.date)
        previous_exercice_start = exercice_start - relativedelta(years=1)
        (
            first_operation_date,
            last_operation_date,
            closure_done,
        ) = self._get_upload_infos_from_operations_stream(exercice_start)
        upload.date = last_operation_date
        cae_accounting_software = get_cae_accounting_software()[0]
        missed_operations: int = 0
        operations = []
        for operation_data in self.producer.stream_operations():
            if cae_accounting_software == "quadra":
                if closure_done and operation_data.date < exercice_start:
                    operation_data.date = exercice_start
                elif operation_data.date < previous_exercice_start:
                    operation_data.date = previous_exercice_start
            company_id = self._find_company_id(operation_data.analytical_account)
            if not company_id:
                missed_operations += 1
            operations.append(
                AccountingOperation(company_id=company_id, **operation_data.__dict__)
            )
        if operations:
            logger.info(f"    {len(operations)} operations")
            upload.operations = operations
        else:
            logger.info(f"    no operation found")
            old_ids = []

        if upload.id:
            upload = DBSESSION().merge(upload)
        else:
            DBSESSION().add(upload)
        DBSESSION().flush()

        return upload.id, missed_operations, old_ids, first_operation_date


def send_error(request, mail_addresses, filename, err):
    """
    Send an error email to mail_addresses
    """
    if mail_addresses:
        try:
            error = getattr(err, "message", str(err))
            message = MAIL_ERROR_BODY.format(error=error, filename=filename)
            subject = MAIL_ERROR_SUBJECT.format(filename=filename)
            send_mail(
                request,
                mail_addresses,
                message,
                subject,
            )
        except Exception:
            logger.exception("send_success error")


def send_unknown_error(request, mail_addresses, filename, err):
    if mail_addresses:
        try:
            subject = MAIL_ERROR_SUBJECT.format(filename=filename)

            message = MAIL_UNKNOWN_ERROR_BODY.format(error=str(err), filename=filename)
            send_mail(
                request,
                mail_addresses,
                message,
                subject,
            )
        except Exception:
            logger.exception("send_success error")


def send_success(request, mail_addresses, filename, new_entries, missing):
    if mail_addresses:
        try:
            subject = MAIL_SUCCESS_SUBJECT.format(filename)
            message = MAIL_SUCCESS_BODY.format(
                filename,
                new_entries,
                missing,
            )
            send_mail(
                request,
                mail_addresses,
                message,
                subject,
            )
        except Exception:
            logger.exception("send_success error")


def _move_file_to_processing(waiting_file: Path):
    """
    MOve the waiting file to the processing queue

    :param waiting_file: The full path to the file to process
    """
    logger.info(" + Moving the file to the processing directory")
    file_to_parse = _mv_file(waiting_file, "processing")
    return file_to_parse


def _is_processing() -> bool:
    """
    Check if there is already a parsing process running
    """
    path = _get_path("processing")

    for filepath in path.iterdir():
        if filepath.is_file():
            return True
    return False


def _prepare_file() -> Optional[Path]:
    """
    Return the path to a file that should be treated (or None if None should be treated)
    """
    pool_path = _get_path("pool")
    waiting_file = _get_file_path_from_pool(pool_path)

    if waiting_file is None:
        return

    already_processing = _is_processing()

    if already_processing:
        logger.info("A parsing is already processing")
        return

    else:
        file_to_parse = _move_file_to_processing(waiting_file)
        logger.info("Parsing an accounting file : %s" % file_to_parse)
        return Path(file_to_parse)


def _get_file_parser(request, file_path: Path) -> BaseParser:
    """
    Return the parser to be used when handling this file
    """
    factory = request.find_service_factory(IAccountingFileParser)
    return factory(file_path, request)


def _get_operation_producer(request, parser: BaseParser) -> BaseProducer:
    """
    Build the tool that produces OperationData
    """
    factory = request.find_service_factory(IAccountingOperationProducer)
    return factory(parser, request)


@celery_app.task(bind=True)
def handle_pool_task(self, force=False):
    """
    Parse the files present in the configured file pool
    """
    # Préparation du fichier à traiter
    file_path = _prepare_file()
    if file_path is None:
        return

    request = get_request()
    mail_addresses = get_recipients_addresses(request)
    try:
        parser = _get_file_parser(request, file_path)
        producer = _get_operation_producer(request, parser)
        handler = AccountingDataHandler(parser, producer, force=force)
        transaction.begin()
        logger.info("  + Storing accounting operations in database")
        (
            upload_object_id,
            missed_associations,
            old_ids,
            first_upload_date,
        ) = handler.run()
        logger.debug("  + File was processed")
        transaction.commit()

    except FileNameException as exc:
        transaction.abort()
        logger.exception("Filename is incorrect")
        logger.exception("* FAILED : transaction has been rollbacked")
        if mail_addresses:
            send_error(request, mail_addresses, file_path.name, exc)
            logger.error("An error mail has been sent to {0}".format(mail_addresses))
        _mv_file(file_path, "error")
        return False

    except Exception as err:
        transaction.abort()
        logger.exception("Unkown Error")
        logger.exception("* FAILED : transaction has been rollbacked")
        if mail_addresses:
            send_unknown_error(request, mail_addresses, file_path.name, err)
            logger.error("An error mail has been sent to {0}".format(mail_addresses))
        _mv_file(file_path, "error")
        logger.error("File has been moved to error directory")
        return False

    else:
        logger.info("Accounting operations where successfully stored")
        _mv_file(file_path)
        logger.info("File has been moved to processed directory")

    if old_ids:
        transaction.begin()
        logger.info("  + Cleaning old operations")
        try:
            _clean_old_operations(old_ids, first_upload_date)
            transaction.commit()
        except Exception:
            transaction.abort()
            logger.exception("Error while cleaning old operations")
        else:
            logger.info(" * Old datas cleaned successfully")

    if upload_object_id:
        compile_measures(upload_object_id)



# -*- coding: utf-8 -*-
import transaction
from celery.utils.log import get_task_logger

from pyramid_celery import celery_app as app

from caerp.models.company import Company

logger = get_task_logger(__name__)


@app.task
def test_task(*args):
    transaction.begin()
    logger.debug("Running the task we want with args : %s" % args)
    company = Company.get(args[0])
    logger.debug(company.name)
    transaction.commit()



import datetime
import logging
import transaction
from pyramid_celery import celery_app

from caerp.models.notification import NotificationEvent
from caerp.utils.notification import (
    publish_event,
    clean_notifications,
)
from caerp_celery.conf import get_request


logger = logging.getLogger(__name__)


@celery_app.task
def publish_pending_notifications_task():
    try:
        request = get_request()
        now = datetime.datetime.now()
        events = NotificationEvent.query().filter(
            NotificationEvent.due_datetime <= now,
            NotificationEvent.published == False,  # noqa:E712
        )
        for event in events:
            if event.is_valid(request):
                publish_event(request, event)
            else:
                request.dbsession.delete(event)
        transaction.commit()
    except Exception:
        logger.exception("Erreur dans publish_pending_notifications_task")
        transaction.abort()


@celery_app.task
def clean_notifications_task():
    """
    Clean notifications in case

    - Notification Event is outdated, conditions are not met anymore
    (e.g : contractor has left)

    - All Notifications have been read
    """
    try:
        request = get_request()
        clean_notifications(request)
        transaction.commit()
    except Exception:
        logger.exception("Erreur dans clean_notification_task")
        transaction.abort()



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst")) as f:
    README = f.read()

with open(os.path.join(here, "requirements.txt")) as f:
    requires = f.read()

with open(os.path.join(here, "CURRENT_VERSION")) as f:
    current_version = f.read().splitlines()[0].strip()


entry_points = {
    "paste.app_factory": [
        "worker = caerp_celery:worker",
        "scheduler = caerp_celery:scheduler",
    ]
}

setup(
    name="caerp_celery",
    version=current_version,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    license="GPLv3",
    description="caerp_celery",
    long_description=README,
    long_description_content_type="text/x-rst",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
    author="Coopérer Pour Entreprendre",
    author_email="contact@endi.coop",
    url="https://framagit.org/caerp/caerp_celery",
    keywords="web wsgi bfg pylons pyramid celery",
    zip_safe=False,
    install_requires=requires,
    entry_points=entry_points,
)



"""
RGPD Related Tasks

1- Collect the data to be cleaned
2- Build the Email
3- 
"""
import logging
import transaction
from pyramid_celery import celery_app

from caerp.models.config import Config
from caerp.consts.rgpd import (
    RGPD_DEFAULT_USERDATA_RETENTION_DAYS, 
    RGPD_DEFAULT_CUSTOMER_RETENTION_DAYS, 
    RGPD_DEFAULT_LOGIN_RETENTION_DAYS,
)
from caerp.services.rgpd.user import (
    get_userdatas_not_used_for,
    get_accounts_not_used_for,
)
from caerp.services.rgpd.customer import get_customers_not_used_for
from caerp.services.rgpd.utils import get_retention_days
from caerp_celery.mail import (
    send_rgpd_userdata_notification,
    send_rgpd_customer_notification,
    send_unused_login_notification,
)
from caerp_celery.conf import get_request, get_sysadmin_mail


logger = logging.getLogger(__name__)


def get_rgpd_email(request):
    return Config.get_value("rgpd_manager_email", default=None, type_=str)


@celery_app.task
def check_rgpd_userdata():
    logger.debug("# RGPD : Check Userdata")
    request = get_request()
    email = get_rgpd_email(request)
    if email is None:
        logger.debug("# RGPD : Aucun responsable RGPD configuré")
        return
    else:
        recipients = [email, get_sysadmin_mail()]
    try:
        request = get_request()
        retention_days = get_retention_days(request, "userdata", RGPD_DEFAULT_USERDATA_RETENTION_DAYS)
        entries = get_userdatas_not_used_for(request, retention_days)
        if len(entries) > 0:
            logger.debug("# {} entrées à nettoyer".format(len(entries)))
            logger.debug("# Envoi d'un mail au responsable")
            send_rgpd_userdata_notification(
                request, recipients, entries, retention_days
            )
    except Exception:
        logger.exception("Erreur lors de la vérification des données de GS à nettoyer")
        transaction.abort()


@celery_app.task
def check_unused_logins():
    logger.debug("# RGPD : Check Unused Logins")
    request = get_request()
    email = get_rgpd_email(request)
    if email is None:
        logger.debug("# RGPD : Aucun responsable RGPD configuré")
        return
    else:
        recipients = [email, get_sysadmin_mail()]
    try:
        request = get_request()
        retention_days = get_retention_days(request, "login", RGPD_DEFAULT_LOGIN_RETENTION_DAYS)
        entries = get_accounts_not_used_for(request, retention_days)
        if len(entries) > 0:
            logger.debug("# {} entrées à nettoyer".format(len(entries)))
            logger.debug("# Envoi d'un mail au responsable")
            send_unused_login_notification(request, recipients, entries, retention_days)
    except Exception:
        logger.exception("Erreur lors de la vérification des comptes à nettoyer")
        transaction.abort()


@celery_app.task
def check_rgpd_customers():
    logger.debug("# RGPD : Check Customers")
    request = get_request()
    email = get_rgpd_email(request)
    if email is None:
        logger.debug("# RGPD : Aucun responsable RGPD configuré")
        return
    else:
        recipients = [email, get_sysadmin_mail()]
    try:
        request = get_request()
        retention_days = get_retention_days(request, "customer", RGPD_DEFAULT_CUSTOMER_RETENTION_DAYS)
        entries = get_customers_not_used_for(request, retention_days)
        if len(entries) > 0:
            logger.debug("# {} entrées à nettoyer".format(len(entries)))
            logger.debug("# Envoi d'un mail au responsable")
            send_rgpd_customer_notification(
                request, recipients, entries, retention_days
            )
    except Exception:
        logger.exception("Erreur lors de la vérification des données client à nettoyer")
        transaction.abort()



"""
Tasks used to compile treasury measures
"""
import datetime
import transaction

from abc import ABC, abstractmethod
from collections import OrderedDict
from dateutil.relativedelta import relativedelta
from pyramid_celery import celery_app
from sqlalchemy import or_, desc, func, extract
from typing import List, Optional, Type, Tuple

from caerp_base.models.base import DBSESSION
from caerp_base.mail import send_mail
from caerp_celery.conf import (
    get_recipients_addresses,
    get_request,
)
from caerp_celery.locks import (
    acquire_lock,
    is_locked,
    release_lock,
)
from caerp_celery.tasks import utils
from caerp.models.accounting.accounting_closures import AccountingClosure
from caerp.models.accounting.balance_sheet_measures import (
    BalanceSheetMeasure,
    BalanceSheetMeasureType,
    BalanceSheetMeasureGrid,
    ActiveBalanceSheetMeasureType,
    PassiveBalanceSheetMeasureType,
)
from caerp.models.accounting.base import (
    BaseAccountingMeasure,
    BaseAccountingMeasureGrid,
    BaseAccountingMeasureType,
    BaseAccountingMeasureTypeCategory,
)
from caerp.models.accounting.income_statement_measures import (
    IncomeStatementMeasure,
    IncomeStatementMeasureGrid,
    IncomeStatementMeasureType,
    IncomeStatementMeasureTypeCategory,
)
from caerp.models.accounting.operations import (
    AccountingOperationUpload,
    AccountingOperation,
)
from caerp.models.accounting.treasury_measures import (
    TreasuryMeasure,
    TreasuryMeasureGrid,
    TreasuryMeasureType,
    TreasuryMeasureTypeCategory,
)
from caerp.models.config import Config


logger = utils.get_logger(__name__)


class GridCacheItem:
    """
    Wraps a grid in cache and provide storages for

    measures
    total value by measure_type
    total value by measure type category
    """

    def __init__(self, grid):
        self.grid = grid
        self._category_totals = {}
        self._type_totals = {}
        self._measures = {}

    def clean_old_measures(self, type_ids):
        """
        Clean old measures

        Set them to  or remove them if theire type is not in the type list

        :param list type_ids: List of ids (int)
        :returns: The list of measures to remove
        """
        old_measures = [
            measure
            for measure in self.grid.measures
            if measure.measure_type_id not in type_ids
        ]
        self.grid.measures = [
            measure
            for measure in self.grid.measures
            if measure.measure_type_id in type_ids
        ]

        for measure in self.grid.measures:
            measure.value = 0
            self._measures[measure.measure_type_id] = measure
        return old_measures

    def get_measures(self):
        """
        Collect all existing measures (newly created or already existing)
        """
        return self._measures

    def store_measure(self, measure):
        """
        Store a measure in the measure cache

        :param obj measure: A child of BaseTreasuryMeasure
        """
        self._measures[measure.measure_type_id] = measure
        if measure not in self.grid.measures:
            self.grid.measures.append(measure)

    def set_total_type(self, measure):
        """
        Store the total value for a given measure type in the grid's cache

        :param obj measure: A child of BaseTreasuryMeasure
        """
        self._type_totals[measure.measure_type_id] = measure.value

    def update_category_total(self, measure):
        """
        Update the category total value for a given measure type in the grid's
        cache

        :param obj measure: A child of BaseTreasuryMeasure
        """
        measure_type = measure.measure_type
        if not measure_type.is_computed_total:
            if measure_type.category_id not in self._category_totals:
                self._category_totals[measure_type.category_id] = 0
            self._category_totals[measure_type.category_id] += measure.value

    def get_type_total(self, typ_id):
        """
        Get the total type for a given measure type

        :param int typ_id: The id of the measure type
        """
        return self._type_totals.get(typ_id, 0)

    def get_category_total(self, category_id):
        """
        Get the category total for a given measure type category

        :param int category_id: The id of the category
        """
        return self._category_totals.get(category_id, 0)


class GridCache:
    """
    A grid cache used to temporarly store all datas related to the current
    computation

    Eache item is an Accounting Measure Grid wraped with the GridCacheItem
    object
    """

    def __init__(self):
        self._datas = {}

    def store_grid(self, key, grid):
        self._datas[key] = GridCacheItem(grid)
        return self._datas[key]

    def get(self, key):
        return self._datas.get(key)


class BaseMeasureCompiler(ABC):
    """
    Base measure compiler

    Uses a cache to store grids currently edited

    For each grid, the cache contains :

        - Measures already created
        - totals by category
        - totals by type

    Totals are used in a second time to compute "total" measures
    (not related to account numbers)
    """

    measure_type_class: BaseAccountingMeasureType
    measure_type_category_class: Optional[BaseAccountingMeasureTypeCategory]
    measure_grid_class: BaseAccountingMeasureGrid
    measure_class: BaseAccountingMeasure
    label: str

    def __init__(self, upload, operations):
        self.upload = upload
        self.filetype = upload.filetype
        self.operations = operations
        self.session = DBSESSION()
        self.cache = GridCache()

        # Collect all types
        self.common_measure_types = self._collect_common_measure_types()
        self.categories = self._collect_categories()
        self.computed_measure_types = self._collect_computed_measure_types()

        self.all_type_ids = [t.id for t in self.common_measure_types]
        for types in self.computed_measure_types.values():
            self.all_type_ids.extend([t.id for t in types])

        self.processed_grids = []

    def _get_active_type_query(self):
        """
        Build a query to list active measure types
        :returns: An sqlalchemy query object
        """
        if self.measure_type_category_class:
            return (
                self.measure_type_class.query()
                .join(self.measure_type_category_class)
                .filter(self.measure_type_class.active == True)  # noqa: E712
                .filter(self.measure_type_category_class.active == True)  # noqa: E712
            )
        else:
            return self.measure_type_class.query().filter(
                self.measure_type_class.active == True
            )  # noqa: E712

    def _collect_common_measure_types(self):
        """
        Retrieve measure types that are not computed from other measures
        :returns: The list of measure types
        """
        return (
            self._get_active_type_query()
            .filter(
                or_(
                    self.measure_type_class.is_total == False,  # noqa: E712
                    self.measure_type_class.total_type == "account_prefix",
                )
            )
            .all()
        )

    def _collect_categories(self) -> Optional[List[BaseAccountingMeasureTypeCategory]]:
        """
        Retrieve all measure type categories
        """
        if self.measure_type_category_class:
            return self.measure_type_category_class.get_categories()
        else:
            return None

    def _collect_computed_measure_types(self):
        """
        Collect computed measure types and store them by category id

        :returns: a dict {category_id: [list of MeasureType]}
        """
        result = OrderedDict()
        query = self._get_active_type_query()

        if self.measure_type_category_class:
            query = (
                query.filter(
                    self.measure_type_class.is_total == True,  # noqa: E712
                )
                .filter(
                    self.measure_type_class.total_type != "account_prefix",
                )
                .order_by(self.measure_type_category_class.order)
                .order_by(self.measure_type_class.order)
            )
        else:
            query = (
                query.filter(
                    self.measure_type_class.is_total == True,  # noqa: E712
                )
                .filter(
                    self.measure_type_class.total_type != "account_prefix",
                )
                .order_by(self.measure_type_class.order)
            )

        for typ in query:
            result.setdefault(typ.category_id, []).append(typ)

        return result

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object

        :param obj grid: The grid instance
        :returns: A valid dict key
        """
        pass

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object

        :param obj operation: The AccountingOperation instance
        :returns: A valid dict key
        """
        pass

    def _get_new_measure(self, measure_type, grid_id):
        """
        Build a new measure
        """
        measure = self.measure_class(
            label=measure_type.label,
            grid_id=grid_id,
            measure_type_id=measure_type.id,
            order=measure_type.order,
        )
        self.session.add(measure)
        self.session.flush()
        return measure

    @abstractmethod
    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load
        specific to the Measure Types

        :param obj grid_item: A GridCacheItem instance
        """
        pass

    @abstractmethod
    def _query_grid(self, operation):
        pass

    @abstractmethod
    def _get_new_grid(self, operation):
        pass

    def get_grid_item(self, operation):
        """
        Retrieve the grid related to the given operation datas

        :param obj operation: an AccountingOperation instance
        :returns: A Grid instance
        """
        key = self.get_cache_key_from_operation(operation)
        grid_item = self.cache.get(key)
        if grid_item is None:
            grid = self._query_grid(operation)
            if grid is None:
                grid = self._get_new_grid(operation)
                self.session.add(grid)
                self.session.flush()
                grid_item = self.cache.store_grid(key, grid)
            else:
                grid_item = self.cache.store_grid(key, grid)
                self._clean_existing_grid(grid_item)
        return grid_item

    def get_values_for_computation(self, grid_item):
        """
        Collect total values in a dict for complex datas computation

        :param obj grid_item: An instance of GridCacheItem
        """
        result = {}
        if self.categories:
            for category in self.categories:
                result[category.label] = grid_item.get_category_total(category.id)

        for typ_ in self.common_measure_types:
            result[typ_.label] = grid_item.get_type_total(typ_.id)

        for typlist_ in list(self.computed_measure_types.values()):
            for typ_ in typlist_:
                value = grid_item.get_type_total(typ_.id)
                result[typ_.label] = value

        return result

    def _cache_grid_totals(self):
        """
        Cache all grid totals in order to use them in computation
        """
        for grid_item in self.processed_grids:
            for measure in list(grid_item.get_measures().values()):
                grid_item.set_total_type(measure)
                grid_item.update_category_total(measure)

    def _process_common_measures(self):
        """
        Compile common measures (related to an account prefix) with the given
        operations

        :returns: A list of GridCacheItem
        """
        logger.debug("    + Processing datas with {}".format(self.__class__.__name__))
        for operation in self.operations:
            if operation.company_id is None:
                continue

            current_grid_item = self.get_grid_item(operation)
            grid = current_grid_item.grid
            if current_grid_item not in self.processed_grids:
                self.processed_grids.append(current_grid_item)

            measures = current_grid_item.get_measures()

            for measure_type in self.common_measure_types:
                if measure_type.match(operation.general_account):
                    measure = measures.get(measure_type.id)
                    if measure is None:
                        measure = self._get_new_measure(
                            measure_type,
                            grid.id,
                        )
                        current_grid_item.store_measure(measure)

                    operation_sign = measure_type.sign()

                    measure.value += operation_sign * operation.total()

                    self.session.merge(measure)
                    self.session.flush()

        return self.processed_grids

    def _complete_grids_with_common_measures(self):
        """
        Insert common measures in the grids if not set yet
        """
        for grid_item in self.processed_grids:
            measures = grid_item.get_measures()
            for measure_type in self.common_measure_types:
                measure = measures.get(measure_type.id)
                if measure is None:
                    measure = self._get_new_measure(
                        measure_type,
                        grid_item.grid.id,
                    )
                    grid_item.store_measure(measure)
                else:
                    measure.order = measure_type.order
                    self.session.merge(measure)

    def _process_computed_measures(self):
        """
        Process dynamically computed measures (based on the other ones)

        :returns: A list of GridCacheItem
        """
        for grid_item in self.processed_grids:
            measures = grid_item.get_measures()
            for category_id, measure_types in list(self.computed_measure_types.items()):
                for measure_type in measure_types:
                    values = self.get_values_for_computation(grid_item)
                    value = measure_type.compute_total(values)
                    measure = measures.get(measure_type.id)
                    if measure is None:
                        measure = self._get_new_measure(
                            measure_type,
                            grid_item.grid.id,
                        )
                        grid_item.store_measure(measure)
                    else:
                        measure.order = measure_type.order

                    measure.value = measure_type.sign() * value
                    self.session.merge(measure)
                    grid_item.set_total_type(measure)

        return self.processed_grids

    def process_datas(self):
        """
        Main entry point to process measure computation
        """
        self._process_common_measures()
        self._complete_grids_with_common_measures()
        self.session.flush()
        self._cache_grid_totals()
        self._process_computed_measures()
        self.session.flush()
        return [griditem.grid for griditem in self.processed_grids]

    def _get_date_to_use_for_grid(self):
        """
        For single grids (BalanceSheet, Treasury), return the date to use for the grid
        """
        if self.upload.filetype == self.upload.SYNCHRONIZED_ACCOUNTING:
            # Remontée auto, on veut la date de mise à jour,
            # pas la date de création (qui date du premier upload, surement
            # le 1er janvier)
            date_to_use = self.upload.updated_at.date()
        else:
            # Cas des fichiers, on veut la date (données du fichier),
            # pas la date de mise à jour (date d'upload)
            date_to_use = self.upload.date
        this_year = datetime.date.today().year

        # si les écritures datent d'années précédentes à l'année en cours,
        if self.upload.date.year < this_year:
            # Only works for 31/12 exercice closure, not a problem as long as
            # synchronized accounting works year by year
            date_to_use = datetime.date(day=31, month=12, year=self.upload.date.year)

        return date_to_use


class BalanceSheetMeasureCompiler(BaseMeasureCompiler):
    measure_type_class = BalanceSheetMeasureType
    measure_type_category_class = None
    measure_grid_class = BalanceSheetMeasureGrid
    measure_class = BalanceSheetMeasure
    label = "Génération des bilans"

    def get_message(self, grids):
        return """Génération des bilans

Bilans générés : {}

""".format(
            len(grids)
        )

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object
        """
        date_to_use = self._get_date_to_use_for_grid()
        key = (operation.company_id, date_to_use)

        return key

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object
        """
        key = (grid.company_id, grid.date)
        return key

    def _query_grid(self, operation):
        """
        Query a grid associated to the given operation

        Query filters should match the get_cache_key_from_operation options
        """
        date_to_use = self._get_date_to_use_for_grid()
        query = (
            BalanceSheetMeasureGrid.query()
            .filter_by(date=date_to_use)
            .filter_by(company_id=operation.company_id)
        )

        query = query.order_by(desc(BalanceSheetMeasureGrid.datetime))
        return query.first()

    def _get_new_grid(self, operation):
        """
        Build a new grid based on the given operation

        :param obj operation: The AccountingOperation from which we build
        measure
        """
        date_to_use = self._get_date_to_use_for_grid()
        return BalanceSheetMeasureGrid(
            date=date_to_use,
            company_id=operation.company_id,
            upload=self.upload,
        )

    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load

        :param obj grid_item: A GridCacheItem instance
        """
        for measure in grid_item.grid.measures:
            self.session.delete(measure)

    def _process_computed_measures(self):
        for grid_item in self.processed_grids:
            measures = grid_item.get_measures()
            for category_id, measure_types in list(self.computed_measure_types.items()):
                for measure_type in measure_types:
                    if measure_type.total_type == "complex_total":
                        values = self._get_values_for_computation(grid_item)
                        value = measure_type.compute_total(values)
                        measure = measures.get(measure_type.id)

                    if measure_type.total_type == "categories":
                        # Total of passive and active (not necessarily meaningfull
                        # in real life)
                        if (
                            measure_type.account_prefix == "active,passive"
                            or measure_type.account_prefix == "passive,active"
                        ):
                            value = self._get_balance_sheet_passive_active_total_value(
                                grid_item, measure_type
                            )
                        elif isinstance(measure_type, ActiveBalanceSheetMeasureType):
                            value = self._get_balance_sheet_total_value(grid_item, True)
                        elif isinstance(measure_type, PassiveBalanceSheetMeasureType):
                            value = self._get_balance_sheet_total_value(
                                grid_item, False
                            )
                        else:
                            raise Exception(
                                "Unexpected measure type (passive or active expected)"
                            )

                    measure = measures.get(measure_type.id)
                    if measure is None:
                        measure = self._get_new_measure(
                            measure_type,
                            grid_item.grid.id,
                        )
                        grid_item.store_measure(measure)
                    else:
                        measure.order = measure_type.order

                    measure.value = measure_type.sign() * value
                    self.session.merge(measure)
                    grid_item.set_total_type(measure)

        return self.processed_grids

    def _get_values_for_computation(self, grid_item):
        """
        Collect total values in a dict for complex datas computation

        :param obj grid_item: An instance of GridCacheItem
        """
        result = {}
        result["Total Actif"] = self._get_balance_sheet_total_value(grid_item, True)
        result["Total Passif"] = self._get_balance_sheet_total_value(grid_item, False)

        for typ_ in self.common_measure_types:
            result[typ_.label] = grid_item.get_type_total(typ_.id)

        for typlist_ in list(self.computed_measure_types.values()):
            for typ_ in typlist_:
                value = grid_item.get_type_total(typ_.id)
                result[typ_.label] = value

        return result

    def _get_balance_sheet_total_value(self, grid_item, active):
        if active:
            polymorphic_id = "active_balance_sheet"
        else:
            polymorphic_id = "passive_balance_sheet"

        query = (
            self.session.query(func.sum(BalanceSheetMeasure.value))
            .filter(BalanceSheetMeasure.grid_id == grid_item.grid.id)
            .outerjoin(BalanceSheetMeasure.measure_type)
            .filter(BalanceSheetMeasureType.type_ == polymorphic_id)
            .filter(BalanceSheetMeasureType.is_total == False)
        )

        if query.first() and query.first()[0] is not None:
            return query.first()[0]
        else:
            return 0

    def _get_balance_sheet_passive_active_total_value(self, grid_item, measure_type):
        query = (
            self.session.query(func.sum(BalanceSheetMeasure.value))
            .filter(BalanceSheetMeasure.grid_id == grid_item.grid.id)
            .outerjoin(BalanceSheetMeasure.measure_type)
            .filter(BalanceSheetMeasureType.is_total == False)
        )

        if query.first() and query.first()[0] is not None:
            return query.first()[0]
        else:
            return 0


class TreasuryMeasureCompiler(BaseMeasureCompiler):
    measure_type_class = TreasuryMeasureType
    measure_type_category_class = TreasuryMeasureTypeCategory
    measure_grid_class = TreasuryMeasureGrid
    measure_class = TreasuryMeasure
    label = "Génération des états de trésorerie"

    def get_message(self, grids):
        return """Génération des états de trésorerie

États de trésorerie générés : {}

""".format(
            len(grids)
        )

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object
        """
        date_to_use = self._get_date_to_use_for_grid()
        key = (operation.company_id, date_to_use)

        return key

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object
        """
        key = (grid.company_id, grid.date)
        return key

    def _query_grid(self, operation):
        """
        Query a grid associated to the given operation

        Query filters should match the get_cache_key_from_operation options
        """
        date_to_use = self._get_date_to_use_for_grid()
        query = (
            TreasuryMeasureGrid.query()
            .filter_by(date=date_to_use)
            .filter_by(company_id=operation.company_id)
        )

        query = query.order_by(desc(TreasuryMeasureGrid.datetime))
        return query.first()

    def _get_new_grid(self, operation):
        """
        Build a new grid based on the given operation

        :param obj operation: The AccountingOperation from which we build
        measure
        """
        date_to_use = self._get_date_to_use_for_grid()
        return TreasuryMeasureGrid(
            date=date_to_use,
            company_id=operation.company_id,
            upload=self.upload,
        )

    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load

        :param obj grid_item: A GridCacheItem instance
        """
        for measure in grid_item.grid.measures:
            self.session.delete(measure)


class IncomeStatementMeasureCompiler(BaseMeasureCompiler):
    measure_type_class = IncomeStatementMeasureType
    measure_type_category_class = IncomeStatementMeasureTypeCategory
    measure_grid_class = IncomeStatementMeasureGrid
    measure_class = IncomeStatementMeasure
    label = "Génération des comptes de résultat"

    def get_message(self, grids):
        return """Génération des comptes de résultat :

Comptes de résultat traités : {}

""".format(
            int(len(grids) / 12)
        )

    def get_cache_key_from_operation(self, operation):
        """
        Build a cache key based on the given operation object
        """
        return (operation.company_id, operation.date.year, operation.date.month)

    def get_cache_key_from_grid(self, grid):
        """
        Build a cache key based on the given grid object
        """
        return (grid.company_id, grid.year, grid.month)

    def _query_grid(self, operation):
        """
        Query a grid associated to the given operation

        Query filters should match the get_cache_key_from_operation options
        """
        return (
            IncomeStatementMeasureGrid.query()
            .filter_by(year=operation.date.year)
            .filter_by(month=operation.date.month)
            .filter_by(company_id=operation.company_id)
            .first()
        )

    def _get_new_grid(self, operation):
        """
        Build a new grid based on the given operation

        :param obj operation: The AccountingOperation from which we build
        measure
        """
        return IncomeStatementMeasureGrid(
            year=operation.date.year,
            month=operation.date.month,
            company_id=operation.company_id,
            upload=self.upload,
        )

    def _clean_existing_grid(self, grid_item):
        """
        Clean an existing grid on first load

        :param obj grid_item: A GridCacheItem instance
        """
        # Only clean the old grid item measures (keep existing)
        old_measures = grid_item.clean_old_measures(self.all_type_ids)
        for measure in old_measures:
            self.session.delete(measure)

    def _process_common_measures(self):
        """
        Add specific process for IncomeStatementMeasureGrid, for now
        setting the updated_at attribute
        """
        logger.debug(
            "  + Processing specific datas for child {}".format(self.__class__.__name__)
        )

        # Calling parent to really process measures
        logger.debug("  + Calling parent process common measures")
        BaseMeasureCompiler._process_common_measures(self)

        # On choisit la date à utiliser comme date de mise à jour des grids
        date_to_set = None
        if self.upload.filetype == AccountingOperationUpload.SYNCHRONIZED_ACCOUNTING:
            date_to_set = self.upload.updated_at
        else:
            # Techniquement c'est la même chose que updated_at
            # mais c'est plus explicite d'utiliser created_at
            date_to_set = self.upload.created_at

        # Setting the date in the non void grids
        for grid in self.processed_grids:
            if grid is not None:
                grid.grid.updated_at = date_to_set
                self.session.merge(grid.grid)

        self.session.flush()
        return self.processed_grids


GRID_COMPILERS = {
    "treasury": TreasuryMeasureCompiler,
    "income_statement": IncomeStatementMeasureCompiler,
    "balance_sheet": BalanceSheetMeasureCompiler,
}


def get_measure_compilers(grid_type=None) -> List[Type[BaseMeasureCompiler]]:
    """
    Retrieve the measure compilers to be used
    """
    if grid_type is not None:
        return [GRID_COMPILERS[grid_type]]
    else:
        return list(GRID_COMPILERS.values())


MAIL_ERROR_SUBJECT = "[Erreur] {}"
MAIL_SUCCESS_SUBJECT = "[Succès] {}"


def send_success(request, mail_addresses, type_label, message):
    """
    Send a success email

    :param obj request: The current request object
    :param list mail_addresses: List of recipients e-mails
    :param str message: message to send
    """
    if mail_addresses:
        try:
            subject = MAIL_SUCCESS_SUBJECT.format(type_label)
            send_mail(request, mail_addresses, message, subject)
        except Exception:
            logger.exception("Error sending email")


def send_error(request, mail_addresses, type_label, message):
    """
    Send an error email

    :param obj request: The current request object
    :param list mail_addresses: List of recipients e-mails
    :param str message: message to send
    """
    if mail_addresses:
        try:
            subject = MAIL_ERROR_SUBJECT.format(type_label)
            send_mail(request, mail_addresses, message, subject)
        except Exception:
            logger.exception("Error sending email")


def get_closure_day_and_month() -> Tuple[int, int]:
    # Get data from caerp's configuration
    closure_day = Config.get_value("accounting_closure_day", default=31, type_=int)
    closure_month = Config.get_value("accounting_closure_month", default=12, type_=int)
    return closure_day, closure_month


def get_exercice_dates_from_date(
    date: datetime.date,
) -> tuple[datetime.date, datetime.date]:
    """
    Return the start and end dates of the fiscal year of the given date
    """
    # Reference year for collecting data
    current_year = date.year
    closure_day, closure_month = get_closure_day_and_month()

    # Find exercice start day and month
    temp_date = datetime.date(
        current_year, closure_month, closure_day
    ) + datetime.timedelta(days=1)
    start_day = temp_date.day
    start_month = temp_date.month

    # Collecting operation for current exercice fiscal
    exercice_start_date = datetime.date(current_year, start_month, start_day)
    exercice_end_date = datetime.date(current_year, closure_month, closure_day)

    # If closure day is not 31/12 we have an exercice overlaping two years
    if not (closure_day == 31 and closure_month == 12):
        # Is the given date after the closure ?
        this_month = date.month
        this_day = date.day
        is_after_the_closure = None
        if this_month > closure_month:
            is_after_the_closure = True
        elif this_month == closure_month:
            if this_day > closure_day:
                is_after_the_closure = True
            else:
                is_after_the_closure = False
        else:
            is_after_the_closure = False

        # If so or not we set the corresponding exercice dates
        if is_after_the_closure:
            exercice_start_date = exercice_start_date.replace(year=current_year)
            exercice_end_date = exercice_end_date + relativedelta(years=1)
        else:
            exercice_start_date = exercice_start_date - relativedelta(years=1)
            exercice_end_date = exercice_end_date.replace(year=current_year)

    return exercice_start_date, exercice_end_date


def collect_operations_by_fiscal_year(
    upload: AccountingOperationUpload, exclude_future_operations: bool = False
) -> List[AccountingOperation]:
    """
    Return the operations needed.
    Collect the operations ("écritures") according to the "exercice fiscal".
    If the "exercice" has been set "closed" in CAErp we don't collect the
    operation of former years (because à-nouveau operation have been done by
    the accountant.
    For now, this function is only for the treasury measures, in the future
    it might also be for the income statement measure.
    """
    logger.info(
        "  + Collecting operations for upload {} (upload_id={})".format(
            upload.date.year, upload.id
        )
    )

    # Get exercice start and end dates
    if upload.date.year == datetime.datetime.now().year:
        ref_date = datetime.datetime.now()
    else:
        ref_date = datetime.date(upload.date.year, 12, 31)
    (exercice_start_date, exercice_end_date) = get_exercice_dates_from_date(ref_date)
    logger.info(
        "  + Exercice starts the {} and ends the {}".format(
            exercice_start_date, exercice_end_date
        )
    )

    # Compute first operation date (through former exercices)
    first_operation_date = exercice_start_date
    first_exercice = (
        DBSESSION().query(extract("year", func.min(AccountingOperation.date))).scalar()
    )
    previous_exercice = exercice_end_date.year - 1
    for exercice in range(previous_exercice, first_exercice - 1, -1):
        exercice_closure = AccountingClosure.query().filter_by(year=exercice).first()
        closure_is_done = exercice_closure.active if exercice_closure else False
        if closure_is_done:
            # Stop on first closure
            break
        else:
            first_operation_date = exercice_start_date - relativedelta(
                years=(exercice_end_date.year - exercice)
            )

    # Compute last operation date (exclude future operations if needed)
    last_operation_date = exercice_end_date
    if exclude_future_operations:
        last_operation_date = min(exercice_end_date, datetime.date.today())

    # Collect operations on computed period
    logger.info(
        "  + Date range is from {} to {}".format(
            first_operation_date, last_operation_date
        )
    )
    collected_operations = (
        AccountingOperation.query()
        .filter(
            AccountingOperation.date.between(first_operation_date, last_operation_date)
        )
        .all()
    )
    logger.info("  + {} operations collected".format(len(collected_operations)))

    return collected_operations


def collect_operations(
    compiler_type: str, upload: AccountingOperationUpload
) -> List[AccountingOperation]:
    """collect operations to use for the given compiler_type

    :param grid_type: _description_
    :type grid_type: str
    :param upload: _description_
    :type upload: AccountingOperationUpload
    :return: _description_
    :rtype: List[AccountingOperation]
    """
    if compiler_type == "income_statement":
        return AccountingOperation.query().filter(
            AccountingOperation.upload_id == upload.id
        )
    elif compiler_type == "balance_sheet":
        return (
            AccountingOperation.query()
            .filter(AccountingOperation.upload_id == upload.id)
            .filter(AccountingOperation.date <= datetime.date.today())
        )
    else:
        return collect_operations_by_fiscal_year(upload, exclude_future_operations=True)


def run_compiler(compiler_type: str, upload_id: int, mail: bool = True):
    """
    Collect accounting operations to be used by the compiler factory
    """
    logger.info(
        f"# Compiling {compiler_type} for AccoutingOperationUpload : {upload_id}"
    )

    messages = []
    transaction.begin()
    upload = AccountingOperationUpload.get(upload_id)
    if upload is None:
        logger.error("  - Upload doesn't exist")
        transaction.abort()
        return False

    operations = collect_operations(compiler_type, upload)
    if not operations:
        logger.error("Can't find any operation")
        transaction.abort()
        return False

    factory = GRID_COMPILERS[compiler_type]
    try:
        compiler = factory(upload, operations)
        grids = compiler.process_datas()
        messages.append(compiler.get_message(grids))
    except Exception as exc:
        logger.exception("Error while generating measures")
        transaction.abort()
        if mail:
            request = get_request()
            mail_addresses = get_recipients_addresses(request)
            message = getattr(exc, "message", str(exc))
            send_error(
                request,
                mail_addresses,
                factory.label,
                message,
            )
        return False
    else:
        logger.info("{0} measure grids were handled".format(len(grids)))
        if mail:
            request = get_request()
            mail_addresses = get_recipients_addresses(request)
            send_success(
                request,
                mail_addresses,
                compiler.label,
                compiler.get_message(grids),
            )
            logger.info("A success email has been sent to {0}".format(mail_addresses))

    transaction.commit()
    logger.info("The transaction has been commited")
    return True


LOCK_NAME = "measure_compute"


def compile_measures(
    upload_id: int, grid_type: Optional[str] = None, mail: bool = True
):
    """Compile accounting measures related to upload id

    :param grid_type: One of the available grid_type (income_statement,
    treasury, balance_sheet)
    """
    logger.info(f"Compiling Accounting States for upload {upload_id}")
    if is_locked(LOCK_NAME):
        logger.error("Other task is running : Cancel")
        return False

    acquire_lock(LOCK_NAME)
    try:
        if grid_type is not None:
            logger.info(f"  + Only compiling {grid_type} grids")
            grid_types = [grid_type]
        else:
            grid_types = list(GRID_COMPILERS.keys())

        for grid_type in grid_types:
            run_compiler(grid_type, upload_id, mail)
    except Exception:
        logger.exception("Erreur inconnue")

    release_lock(LOCK_NAME)
    return True


@celery_app.task(bind=True)
def compile_measures_task(self, upload_id, grid_type=None):
    """
    Celery task handling measures compilation
    """
    logger.info("# Launching the compile measure task #")
    res = compile_measures(upload_id, grid_type)
    if res:
        logger.info("===> Task Finished")


@celery_app.task(bind=True)
def scheduled_compile_measure_task(self, force=False):
    """
    Scheduled Celery task to handle automatic measures compilation
    for Synchronized Datas

    ACTIVATE THIS TASK IN CELERY'S INI CONFIG FILE IF NEEDED
    Eg:
        [celerybeat:accounting_measure_compute]
        task = caerp_celery.tasks.accounting_measure_compute.scheduled_compile_measure_task  # NOQA: 905
        type = crontab
        schedule = {"minute": 0, "hour": 6}
    """
    logger.info("# Launching scheduled compile measure task for synchronized datas #")
    yesterday = datetime.date.today() - datetime.timedelta(1)
    uploads = (
        AccountingOperationUpload.query()
        .filter(
            AccountingOperationUpload.filetype
            == AccountingOperationUpload.SYNCHRONIZED_ACCOUNTING
        )
        .filter(AccountingOperationUpload.updated_at >= yesterday)
        .order_by(AccountingOperationUpload.date)
        .all()
    )
    if not uploads:
        logger.error(
            "ABORT : no upload with 'synchronized_accounting' type \
since yesterday"
        )
    else:
        for upload in uploads:
            compile_measures(upload.id, mail=False)
        logger.info("===> End of scheduled compile measure task for synchronized datas")



import transaction

from caerp.export.task_pdf import ensure_task_pdf_persisted
from caerp_celery.conf import get_request
from caerp_celery.hacks import setup_rendering_hacks
from caerp_celery.mail import (
    send_customer_new_order_mail,
    send_supplier_new_order_mail,
    send_customer_new_invoice_mail,
    send_supplier_new_invoice_mail,
)
from caerp_celery.tasks import utils
from caerp_celery.transactional_task import task_tm


logger = utils.get_logger(__name__)


def _get_task_from_id(task_id):
    document = utils.get_task(task_id)
    if document is None:
        raise Exception(f"Document with id {task_id} doesn't exist in database")
    return document


@task_tm
def scheduled_render_pdf_task(document_id):
    logger.debug("Scheduling a PDF render Task for {}".format(document_id))
    document = _get_task_from_id(document_id)
    request = get_request()
    try:
        setup_rendering_hacks(request, document)
        utils.set_current_user(request, document.status_user_id)
        ensure_task_pdf_persisted(document, request)
        transaction.commit()
    except Exception:
        logger.exception("Error in scheduled_render_pdf_task")
        transaction.abort()


@task_tm
def async_internalestimation_valid_callback(document_id):
    """
    Handle the transfer of an InternalEstimation to the Client Company

    - Ensure supplier exists
    - Generates the PDF
    - Create a Supplier Order and attache the pdf file
    """
    logger.debug(f"Async internal estimation validation callback for {document_id}")
    document = _get_task_from_id(document_id)
    request = get_request()
    try:
        logger.debug("Setup rendering hacks")
        setup_rendering_hacks(request, document)
        utils.set_current_user(request, document.status_user_id)
        order = document.sync_with_customer(request)
        send_customer_new_order_mail(request, order)
        send_supplier_new_order_mail(request, order)
        transaction.commit()
    except Exception:
        logger.exception("Error in async_internalestimation_valid_callback")
        transaction.abort()


@task_tm
def async_internalinvoice_valid_callback(document_id):
    """
    Handle the transfer of an InternalInvoice to the Client Company

    - Ensure supplier exists
    - Generates the PDF
    - Create a Supplier Invoice and attach the pdf file
    """
    logger.debug(f"Async internal invoice validation callback for {document_id}")
    document = _get_task_from_id(document_id)
    request = get_request()
    try:
        setup_rendering_hacks(request, document)
        utils.set_current_user(request, document.status_user_id)
        supplier_invoice = document.sync_with_customer(request)
        send_customer_new_invoice_mail(request, supplier_invoice)
        send_supplier_new_invoice_mail(request, supplier_invoice)
        transaction.commit()
    except Exception:
        logger.exception("Error in async_internalinvoice_valid_callback")
        transaction.abort()
